from __future__ import annotations

import itertools
from datetime import datetime
from typing import Any, Dict, Iterable, Iterator, List, Optional, Text

import structlog
from pymongo.synchronous.collection import Collection

from rasa.core.brokers.broker import EventBroker
from rasa.core.tracker_stores.tracker_store import SerializedTrackerAsText, TrackerStore
from rasa.shared.core.domain import Domain
from rasa.shared.core.events import SessionStarted
from rasa.shared.core.trackers import DialogueStateTracker, EventVerbosity

structlogger = structlog.get_logger(__name__)


class MongoTrackerStore(TrackerStore, SerializedTrackerAsText):
    """Stores conversation history in Mongo.

    Property methods:
        conversations: returns the current conversation
    """

    def __init__(
        self,
        domain: Domain,
        host: Optional[Text] = "mongodb://localhost:27017",
        db: Optional[Text] = "rasa",
        username: Optional[Text] = None,
        password: Optional[Text] = None,
        auth_source: Optional[Text] = "admin",
        collection: Text = "conversations",
        event_broker: Optional[EventBroker] = None,
        **kwargs: Dict[Text, Any],
    ) -> None:
        from pymongo import MongoClient
        from pymongo.database import Database

        self.client: MongoClient = MongoClient(
            host,
            username=username,
            password=password,
            authSource=auth_source,
            # delay connect until process forking is done
            connect=False,
        )

        self.db = Database(self.client, db)
        self.collection = collection
        super().__init__(domain, event_broker, **kwargs)

        self._ensure_indices()

    @property
    def conversations(self) -> Collection:
        """Returns the current conversation."""
        return self.db[self.collection]

    def _ensure_indices(self) -> None:
        """Create an index on the sender_id."""
        self.conversations.create_index("sender_id")

    @staticmethod
    def _current_tracker_state_without_events(tracker: DialogueStateTracker) -> Dict:
        # get current tracker state and remove `events` key from state
        # since events are pushed separately in the `update_one()` operation
        state = tracker.current_state(EventVerbosity.ALL)
        state.pop("events", None)

        return state

    async def delete(self, sender_id: Text) -> None:
        """Delete tracker for the given sender_id.

        Args:
            sender_id: Sender id of the tracker to be deleted.
        """
        if not await self.exists(sender_id):
            structlogger.info(
                "mongo_tracker_store.delete.no_tracker_for_sender_id",
                event_info=f"Could not find tracker for conversation ID '{sender_id}'.",
            )
            return None
        self.conversations.delete_one({"sender_id": sender_id})

        structlogger.info(
            "mongo_tracker_store.delete.deleted_tracker",
            sender_id=sender_id,
        )

    async def save(self, tracker: DialogueStateTracker) -> None:
        """Saves the current conversation state."""
        await self.stream_events(tracker)

        additional_events = self._additional_events(tracker)

        self.conversations.update_one(
            {"sender_id": tracker.sender_id},
            {
                "$set": self._current_tracker_state_without_events(tracker),
                "$push": {
                    "events": {"$each": [e.as_dict() for e in additional_events]}
                },
            },
            upsert=True,
        )

    def _additional_events(self, tracker: DialogueStateTracker) -> Iterator:
        """Return events from the tracker which aren't currently stored.

        Args:
            tracker: Tracker to inspect.

        Returns:
            List of serialised events that aren't currently stored.

        """
        stored = self.conversations.find_one({"sender_id": tracker.sender_id}) or {}
        all_events = self._events_from_serialized_tracker(stored)

        number_events_since_last_session = len(
            self._events_since_last_session_start(all_events)
        )

        return itertools.islice(
            tracker.events, number_events_since_last_session, len(tracker.events)
        )

    @staticmethod
    def _events_from_serialized_tracker(serialised: Dict) -> List[Dict]:
        return serialised.get("events", [])

    @staticmethod
    def _events_since_last_session_start(events: List[Dict]) -> List[Dict]:
        """Retrieve events since and including the latest `SessionStart` event.

        Args:
            events: All events for a conversation ID.

        Returns:
            List of serialised events since and including the latest `SessionStarted`
            event. Returns all events if no such event is found.

        """
        events_after_session_start = []
        for event in reversed(events):
            events_after_session_start.append(event)
            if event["event"] == SessionStarted.type_name:
                break

        return list(reversed(events_after_session_start))

    async def _retrieve(
        self, sender_id: Text, fetch_events_from_all_sessions: bool
    ) -> Optional[List[Dict[Text, Any]]]:
        stored = self.conversations.find_one({"sender_id": sender_id})

        # look for conversations which have used an `int` sender_id in the past
        # and update them.
        if not stored and sender_id.isdigit():
            from pymongo import ReturnDocument

            stored = self.conversations.find_one_and_update(
                {"sender_id": int(sender_id)},
                {"$set": {"sender_id": str(sender_id)}},
                return_document=ReturnDocument.AFTER,
            )

        if not stored:
            return None

        events = self._events_from_serialized_tracker(stored)

        if not fetch_events_from_all_sessions:
            events = self._events_since_last_session_start(events)

        return events

    async def retrieve(self, sender_id: Text) -> Optional[DialogueStateTracker]:
        """Retrieves tracker for the latest conversation session."""
        events = await self._retrieve(sender_id, fetch_events_from_all_sessions=False)

        if not events:
            return None

        return DialogueStateTracker.from_dict(sender_id, events, self.domain.slots)

    async def retrieve_full_tracker(
        self, conversation_id: Text
    ) -> Optional[DialogueStateTracker]:
        """Fetching all tracker events across conversation sessions."""
        events = await self._retrieve(
            conversation_id, fetch_events_from_all_sessions=True
        )

        if not events:
            return None

        return DialogueStateTracker.from_dict(
            conversation_id, events, self.domain.slots
        )

    async def keys(self) -> Iterable[Text]:
        """Returns sender_ids of the Mongo Tracker Store."""
        return [c["sender_id"] for c in self.conversations.find()]

    async def update(self, tracker: DialogueStateTracker) -> None:
        """Overwrites the tracker for the given sender_id."""
        self.conversations.replace_one(
            {"sender_id": tracker.sender_id},
            tracker.current_state(EventVerbosity.ALL),
            upsert=True,
        )

        first_event_timestamp = str(datetime.fromtimestamp(tracker.events[0].timestamp))

        structlogger.info(
            "redis_tracker_store.update.updated_tracker",
            sender_id=tracker.sender_id,
            first_event_timestamp=first_event_timestamp,
        )



from __future__ import annotations

import audioop
import json
import uuid
from typing import Any, Awaitable, Callable, Dict, Optional, Text, Tuple

import structlog
from sanic import (  # type: ignore[attr-defined]
    Blueprint,
    HTTPResponse,
    Request,
    Websocket,
    response,
)

from rasa.core.channels import UserMessage, requires_basic_auth
from rasa.core.channels.voice_ready.utils import (
    CallParameters,
    validate_username_password_credentials,
)
from rasa.core.channels.voice_stream.audio_bytes import RasaAudioBytes
from rasa.core.channels.voice_stream.call_state import call_state
from rasa.core.channels.voice_stream.tts.tts_engine import TTSEngine
from rasa.core.channels.voice_stream.util import repack_voice_credentials
from rasa.core.channels.voice_stream.voice_channel import (
    ContinueConversationAction,
    DTMFInputAction,
    EndConversationAction,
    NewAudioAction,
    VoiceChannelAction,
    VoiceInputChannel,
    VoiceOutputChannel,
)

logger = structlog.get_logger()

JAMBONZ_STREAMS_WEBSOCKET_PATH = "webhooks/jambonz_stream/websocket"


def map_call_params(data: Dict[Text, str]) -> CallParameters:
    """Map the Jambonz stream parameters to the CallParameters dataclass."""
    call_sid = data.get("callSid", "None")
    from_number = data.get("from", "Unknown")
    to_number = data.get("to")
    return CallParameters(
        call_id=call_sid,
        user_phone=from_number,
        bot_phone=to_number,
        stream_id=call_sid,
    )


class JambonzStreamOutputChannel(VoiceOutputChannel):
    @classmethod
    def name(cls) -> str:
        return "jambonz_stream"

    async def send_audio_bytes(
        self, recipient_id: str, audio_bytes: RasaAudioBytes
    ) -> None:
        """Overridden to send binary websocket messages for Jambonz.

        Converts 8kHz μ-law to 8kHz L16 PCM for Jambonz streaming.
        """
        pcm = audioop.ulaw2lin(audio_bytes, 2)
        await self.voice_websocket.send(pcm)

    def create_marker_message(self, recipient_id: str) -> Tuple[str, str]:
        """Create a marker message to track audio stream position."""
        marker_id = uuid.uuid4().hex
        return json.dumps({"type": "mark", "data": {"name": marker_id}}), marker_id


class JambonzStreamInputChannel(VoiceInputChannel):
    @classmethod
    def name(cls) -> str:
        return "jambonz_stream"

    def __init__(
        self,
        server_url: str,
        asr_config: Dict,
        tts_config: Dict,
        interruptions: Optional[Dict[str, int]] = None,
        username: Optional[Text] = None,
        password: Optional[Text] = None,
    ) -> None:
        """Initialize the channel.

        Args:
            username: Optional username for basic auth
            password: Optional password for basic auth
        """
        super().__init__(server_url, asr_config, tts_config, interruptions)
        self.username = username
        self.password = password

    @classmethod
    def from_credentials(
        cls, credentials: Optional[Dict[Text, Any]]
    ) -> JambonzStreamInputChannel:
        """Create a channel from credentials dictionary.

        Args:
            credentials: Dictionary containing the required credentials:
                - server_url: URL where the server is hosted
                - asr: ASR engine configuration
                - tts: TTS engine configuration
                - username: Optional username for basic auth
                - password: Optional password for basic auth

        Returns:
            JambonzStreamInputChannel instance
        """
        # Get common credentials from parent
        cls.validate_credentials(credentials)
        new_creds = repack_voice_credentials(credentials)
        return cls(**new_creds)

    @classmethod
    def validate_credentials(cls, credentials: Optional[Dict[Text, Any]]) -> None:
        cls.validate_basic_credentials(credentials)
        # Check optional basic auth credentials
        username = credentials.get("username")  # type: ignore[union-attr]
        password = credentials.get("password")  # type: ignore[union-attr]
        validate_username_password_credentials(username, password, "Jambonz Stream")

    def _websocket_stream_url(self) -> str:
        """Returns the websocket stream URL."""
        # depending on the config value, the url might contain http as a
        # protocol or not - we'll make sure both work
        if self.server_url.startswith("http"):
            base_url = self.server_url.replace("http", "ws")
        else:
            base_url = f"wss://{self.server_url}"
        return f"{base_url}/{JAMBONZ_STREAMS_WEBSOCKET_PATH}"

    def channel_bytes_to_rasa_audio_bytes(self, input_bytes: bytes) -> RasaAudioBytes:
        """Convert Jambonz audio bytes (L16 PCM) to Rasa audio bytes (μ-law)."""
        ulaw = audioop.lin2ulaw(input_bytes, 2)
        return RasaAudioBytes(ulaw)

    async def collect_call_parameters(
        self, channel_websocket: Websocket
    ) -> Optional[CallParameters]:
        # Wait for initial metadata message
        message = await channel_websocket.recv()
        logger.debug("jambonz.collect_call_parameters", message=message)
        metadata = json.loads(message)
        return map_call_params(metadata)

    def map_input_message(self, message: Any, ws: Websocket) -> VoiceChannelAction:
        # Handle binary audio frames
        if isinstance(message, bytes):
            channel_bytes = message
            audio_bytes = self.channel_bytes_to_rasa_audio_bytes(channel_bytes)
            return NewAudioAction(audio_bytes)

        # Handle JSON messages
        data = json.loads(message)
        if data.get("event") == "dtmf":
            return DTMFInputAction(
                digit=data["dtmf"],
            )
        if data.get("type") == "mark":
            if data["data"]["name"] == call_state.latest_bot_audio_id:
                # Just finished streaming last audio bytes
                call_state.is_bot_speaking = False
                if call_state.should_hangup:
                    logger.debug(
                        "jambonz.hangup", marker=call_state.latest_bot_audio_id
                    )
                    return EndConversationAction()
            else:
                call_state.is_bot_speaking = True
        else:
            logger.warning("jambonz.unexpected_message", message=data)

        return ContinueConversationAction()

    def create_output_channel(
        self, voice_websocket: Websocket, tts_engine: TTSEngine
    ) -> VoiceOutputChannel:
        return JambonzStreamOutputChannel(
            voice_websocket,
            tts_engine,
            self.tts_cache,
        )

    async def interrupt_playback(
        self, ws: Websocket, call_parameters: CallParameters
    ) -> None:
        """Interrupt the current playback of audio."""
        logger.debug("jambonz.interrupt_playback")
        await ws.send(json.dumps({"type": "killAudio"}))

    def blueprint(
        self, on_new_message: Callable[[UserMessage], Awaitable[Any]]
    ) -> Blueprint:
        blueprint = Blueprint("jambonz_stream", __name__)

        @blueprint.route("/", methods=["GET"])
        async def health(_: Request) -> HTTPResponse:
            return response.json({"status": "ok"})

        @blueprint.route("/call_status", methods=["POST"])
        @requires_basic_auth(self.username, self.password)
        async def call_status(request: Request) -> HTTPResponse:
            """Handle call status updates from Jambonz."""
            data = request.json
            logger.debug("jambonz.call_status.received", data=data)
            return response.json({"status": "ok"})

        @blueprint.route("/webhook", methods=["POST"])
        @requires_basic_auth(self.username, self.password)
        async def webhook(request: Request) -> HTTPResponse:
            """Handle incoming webhook requests from Jambonz."""
            data = request.json
            logger.debug("jambonz.webhook.received", data=data)
            return response.json(
                [
                    {
                        "verb": "listen",
                        "url": self._websocket_stream_url(),
                        "sampleRate": 8000,
                        "passDtmf": True,
                        "bidirectionalAudio": {
                            "enabled": True,
                            "streaming": True,
                            "sampleRate": 8000,
                        },
                    }
                ]
            )

        @blueprint.websocket("/websocket", subprotocols=["audio.jambonz.org"])  # type: ignore[misc]
        async def handle_message(request: Request, ws: Websocket) -> None:
            try:
                await self.run_audio_streaming(on_new_message, ws)
            except Exception as e:
                logger.error("jambonz.handle_message.error", error=e)

        return blueprint



import os
import shutil
import tempfile
from typing import Any, List

from pydantic import BaseModel

from rasa.nlu.utils import write_json_to_file
from rasa.shared.utils.io import read_json_file

ORIGIN_DB_PATH = "db"
CONTACTS = "contacts.json"
TRANSACTIONS = "transactions.json"
MY_ACCOUNT = "my_account.json"
CARDS = "cards.json"


class MyAccount(BaseModel):
    account: str
    funds: int
    daily_limit: int
    monthly_limit: int


class Transaction(BaseModel):
    datetime: str
    recipient: str
    sender: str
    amount: str
    description: str

    def stringify(self) -> str:
        from datetime import datetime

        parsed_date = datetime.strptime(self.datetime, "%Y-%m-%dT%H:%M:%S.%f")
        formatted_date = parsed_date.strftime("%Y-%m-%d %H:%M:%S")

        return (
            f"- {self.amount} from {self.sender} to "
            f"{self.recipient}: {self.description} ({formatted_date})"
        )


class Contact(BaseModel):
    name: str
    handle: str


class Card(BaseModel):
    name: str
    number: str
    expiration_date: str
    cvv: str
    handle: str
    status: str


def get_session_db_path(session_id: str) -> str:
    tempdir = tempfile.gettempdir()
    project_name = "rasa-calm-demo"
    return os.path.join(tempdir, project_name, session_id)


def prepare_db_file(session_id: str, db: str) -> str:
    session_db_path = get_session_db_path(session_id)
    os.makedirs(session_db_path, exist_ok=True)
    destination_file = os.path.join(session_db_path, db)
    if not os.path.exists(destination_file):
        origin_file = os.path.join(ORIGIN_DB_PATH, db)
        shutil.copy(origin_file, destination_file)
    return destination_file


def read_db(session_id: str, db: str) -> Any:
    db_file = prepare_db_file(session_id, db)
    return read_json_file(db_file)


def write_db(session_id: str, db: str, data: Any) -> None:
    db_file = prepare_db_file(session_id, db)
    write_json_to_file(db_file, data)


def get_contacts(session_id: str) -> List[Contact]:
    return [Contact(**item) for item in read_db(session_id, CONTACTS)]


def get_cards(session_id: str) -> List[Card]:
    return [Card(**item) for item in read_db(session_id, CARDS)]


def get_transactions(session_id: str) -> List[Transaction]:
    return [Transaction(**item) for item in read_db(session_id, TRANSACTIONS)]


def get_account(session_id: str) -> MyAccount:
    return MyAccount(**read_db(session_id, MY_ACCOUNT))


def write_account(session_id: str, account: MyAccount) -> None:
    write_db(session_id, MY_ACCOUNT, account.dict())


def add_contact(session_id: str, contact: Contact) -> None:
    contacts = get_contacts(session_id)
    contacts.append(contact)
    write_db(session_id, CONTACTS, [c.dict() for c in contacts])


def add_transaction(session_id: str, transaction: Transaction) -> None:
    transactions = get_transactions(session_id)
    transactions.append(transaction)
    write_db(session_id, TRANSACTIONS, [t.dict() for t in transactions])


def write_contacts(session_id: str, contacts: List[Contact]) -> None:
    write_db(session_id, CONTACTS, [c.dict() for c in contacts])



import os
from enum import Enum
from typing import Any, Dict, List, Optional, Text, Tuple, Union

import redis
import structlog
from pydantic import BaseModel, ConfigDict

from rasa.core.constants import (
    AWS_ELASTICACHE_CLUSTER_NAME_ENV_VAR_NAME,
    REDIS_SERVICE_NAME,
)
from rasa.core.iam_credentials_providers.credentials_provider_protocol import (
    IAMCredentialsProvider,
    IAMCredentialsProviderInput,
    SupportedServiceType,
    create_iam_credentials_provider,
)
from rasa.shared.exceptions import ConnectionException, RasaException

structlogger = structlog.getLogger(__name__)

DEFAULT_SOCKET_TIMEOUT_IN_SECONDS = 10


class DeploymentMode(Enum):
    """Supported Redis deployment modes."""

    STANDARD = "standard"
    CLUSTER = "cluster"
    SENTINEL = "sentinel"


class StandardRedisConfig(BaseModel):
    """Base configuration for Redis connections."""

    model_config = ConfigDict(arbitrary_types_allowed=True)

    host: Text = "localhost"
    port: int = 6379
    username: Optional[Text] = None
    password: Optional[Text] = None
    use_ssl: bool = False
    ssl_keyfile: Optional[Text] = None
    ssl_certfile: Optional[Text] = None
    ssl_ca_certs: Optional[Text] = None
    db: int = 0
    socket_timeout: float = DEFAULT_SOCKET_TIMEOUT_IN_SECONDS
    decode_responses: bool = False
    iam_credentials_provider: Optional[IAMCredentialsProvider] = None


class ClusterRedisConfig(StandardRedisConfig):
    """Configuration for Redis Cluster connections."""

    endpoints: List[Tuple[Text, int]]


class SentinelRedisConfig(StandardRedisConfig):
    """Configuration for Redis Sentinel connections."""

    endpoints: List[Tuple[Text, int]]
    sentinel_service: Optional[Text] = "mymaster"


class RedisConfig(BaseModel):
    """Base configuration for Redis connections."""

    host: Text = "localhost"
    port: int = 6379
    service_type: SupportedServiceType
    username: Optional[Text] = None
    password: Optional[Text] = None
    use_ssl: bool = False
    ssl_keyfile: Optional[Text] = None
    ssl_certfile: Optional[Text] = None
    ssl_ca_certs: Optional[Text] = None
    db: int = 0
    socket_timeout: float = DEFAULT_SOCKET_TIMEOUT_IN_SECONDS
    decode_responses: bool = False
    deployment_mode: Text = DeploymentMode.STANDARD.value
    endpoints: Optional[List[Text]] = None
    sentinel_service: Optional[Text] = None


class RedisConnectionFactory:
    """Factory class for creating Redis connections with different modes."""

    @classmethod
    def create_connection(
        cls,
        config: RedisConfig,
    ) -> Union[redis.Redis, redis.RedisCluster]:
        """Create a Redis connection based on the configuration.

        Args:
            config: Redis configuration object containing all connection parameters.

        Returns:
            A Redis connection - either a standard Redis connection, a RedisCluster
            connection, or a Redis master connection managed by Sentinel.

        Raises:
            RasaException: If configuration is invalid.
        """
        if config.endpoints is None:
            config.endpoints = []

        try:
            deployment_mode_enum = DeploymentMode(config.deployment_mode)
        except ValueError:
            valid_modes = [mode.value for mode in DeploymentMode]
            raise RasaException(
                f"Invalid deployment_mode '{config.deployment_mode}'. "
                f"Must be one of: {', '.join(valid_modes)}"
            )

        parsed_endpoints = cls._parse_and_validate_endpoints(
            deployment_mode_enum, config.endpoints, config.host, config.port
        )

        iam_credentials_provider = create_iam_credentials_provider(
            IAMCredentialsProviderInput(
                service_type=config.service_type,
                service_name=REDIS_SERVICE_NAME,
                username=config.username,
                cluster_name=os.getenv(AWS_ELASTICACHE_CLUSTER_NAME_ENV_VAR_NAME),
            )
        )

        if deployment_mode_enum == DeploymentMode.CLUSTER:
            cls._log_cluster_db_warning(deployment_mode_enum, config.db)
            cluster_config = ClusterRedisConfig(
                username=config.username,
                password=config.password,
                use_ssl=config.use_ssl,
                ssl_certfile=config.ssl_certfile,
                ssl_keyfile=config.ssl_keyfile,
                ssl_ca_certs=config.ssl_ca_certs,
                db=config.db,
                socket_timeout=config.socket_timeout,
                decode_responses=config.decode_responses,
                endpoints=parsed_endpoints,
                iam_credentials_provider=iam_credentials_provider,
            )
            return cls._create_cluster_connection(cluster_config)
        elif deployment_mode_enum == DeploymentMode.SENTINEL:
            sentinel_kwargs = {
                "username": config.username,
                "password": config.password,
                "use_ssl": config.use_ssl,
                "ssl_certfile": config.ssl_certfile,
                "ssl_keyfile": config.ssl_keyfile,
                "ssl_ca_certs": config.ssl_ca_certs,
                "db": config.db,
                "socket_timeout": config.socket_timeout,
                "decode_responses": config.decode_responses,
                "endpoints": parsed_endpoints,
                "iam_credentials_provider": iam_credentials_provider,
            }

            if config.sentinel_service is not None:
                sentinel_kwargs["sentinel_service"] = config.sentinel_service

            sentinel_config = SentinelRedisConfig(**sentinel_kwargs)
            return cls._create_sentinel_connection(sentinel_config)
        else:
            standard_config = StandardRedisConfig(
                host=config.host,
                port=config.port,
                username=config.username,
                password=config.password,
                use_ssl=config.use_ssl,
                ssl_certfile=config.ssl_certfile,
                ssl_keyfile=config.ssl_keyfile,
                ssl_ca_certs=config.ssl_ca_certs,
                db=config.db,
                socket_timeout=config.socket_timeout,
                decode_responses=config.decode_responses,
                iam_credentials_provider=iam_credentials_provider,
            )
            return cls._create_standard_connection(standard_config)

    @classmethod
    def _parse_and_validate_endpoints(
        cls,
        deployment_mode: DeploymentMode,
        endpoints: List[str],
        host: Text,
        port: int,
    ) -> List[Union[Dict[str, Any], Tuple[str, int]]]:
        """Parse and validate endpoints based on deployment mode."""
        if deployment_mode == DeploymentMode.STANDARD:
            if endpoints:
                structlogger.warning(
                    "redis_connection_factory.standard_endpoints_ignored",
                    event_info="Parameter `endpoints` ignored in standard mode. "
                    "Only 'host' and 'port' are used for standard Redis connections.",
                )
            return []

        if not endpoints:
            endpoints = cls._get_default_endpoints(deployment_mode, host, port)

        # Parse endpoints into appropriate format
        parsed_endpoints = cls._parse_all_endpoints(endpoints, deployment_mode)

        if not parsed_endpoints:
            raise RasaException(
                f"No valid '{deployment_mode.value}' endpoints provided"
            )

        return parsed_endpoints

    @classmethod
    def _get_default_endpoints(
        cls,
        deployment_mode: DeploymentMode,
        host: Text,
        port: int,
    ) -> List[str]:
        """Get default endpoints when none provided."""
        if deployment_mode == DeploymentMode.CLUSTER:
            structlogger.warning(
                "redis_connection_factory.cluster_endpoints_not_provided",
                event_info="No endpoints provided for cluster mode. "
                "Using default 'host:port' configuration.",
            )
            return [f"{host}:{port}"]
        elif deployment_mode == DeploymentMode.SENTINEL:
            raise RasaException("Sentinel mode requires endpoints configuration")

        return []

    @classmethod
    def _parse_single_endpoint(
        cls, endpoint: str, deployment_mode: DeploymentMode
    ) -> Optional[Union[Dict[str, Any], Tuple[str, int]]]:
        """Parse a single endpoint string into the appropriate format."""
        if not isinstance(endpoint, str):
            structlogger.warning(
                f"redis_connection_factory.invalid_{deployment_mode.value}_endpoint_type",
                event_info=f"Invalid endpoint type for endpoint '{endpoint}'. "
                "Expected string in 'host:port' format.",
            )
            return None

        if ":" not in endpoint:
            structlogger.warning(
                f"redis_connection_factory.invalid_{deployment_mode.value}_endpoint_format",
                event_info=f"Invalid format for endpoint '{endpoint}'. "
                "Expected 'host:port'.",
            )
            return None

        host, port_str = endpoint.rsplit(":", 1)
        try:
            port = int(port_str)
        except ValueError:
            structlogger.warning(
                f"redis_connection_factory.invalid_{deployment_mode.value}_endpoint",
                event_info=f"Invalid port in endpoint '{endpoint}'. "
                "Expected format 'host:port'.",
            )
            return None

        return (host, port)

    @classmethod
    def _parse_all_endpoints(
        cls,
        endpoints: List[str],
        deployment_mode: DeploymentMode,
    ) -> List[Union[Dict[str, Any], Tuple[str, int]]]:
        """Parse a list of endpoint strings into appropriate format."""
        parsed_endpoints = []
        for endpoint in endpoints:
            parsed = cls._parse_single_endpoint(endpoint, deployment_mode)
            if parsed:
                parsed_endpoints.append(parsed)
        return parsed_endpoints

    @classmethod
    def _create_cluster_connection(
        cls,
        config: ClusterRedisConfig,
    ) -> redis.RedisCluster:
        """Create a Redis Cluster connection.

        Note: Database parameter is ignored in cluster mode (always uses db=0).

        Args:
            config: Cluster configuration containing all connection parameters.

        Returns:
            redis.RedisCluster: Configured cluster connection.

        Raises:
            ConnectionException: If cluster initialization fails.
        """
        from redis.cluster import ClusterNode

        structlogger.info(
            "redis_connection_factory.cluster_mode",
            event_info=f"Initializing Redis Cluster with {len(config.endpoints)} nodes",
        )

        cluster_nodes = [ClusterNode(host, port) for host, port in config.endpoints]

        common_config_kwargs = {
            "startup_nodes": cluster_nodes,
            "ssl": config.use_ssl,
            "ssl_certfile": config.ssl_certfile,
            "ssl_keyfile": config.ssl_keyfile,
            "ssl_ca_certs": config.ssl_ca_certs,
            "socket_timeout": config.socket_timeout,
            "decode_responses": config.decode_responses,
        }

        try:
            if config.iam_credentials_provider is not None:
                structlogger.debug("redis_connection_factory.cluster_iam_auth_enabled")

                redis_cluster: redis.RedisCluster = redis.RedisCluster(
                    credential_provider=config.iam_credentials_provider,
                    **common_config_kwargs,
                )
            else:
                redis_cluster = redis.RedisCluster(
                    username=config.username,
                    password=config.password,
                    **common_config_kwargs,
                )
        except Exception as e:
            raise ConnectionException(f"Error initializing Redis Cluster: {e}")

        return redis_cluster

    @classmethod
    def _create_sentinel_connection(
        cls,
        config: SentinelRedisConfig,
    ) -> redis.Redis:
        """Create a Sentinel-managed Redis connection.

        Connects to Redis master through Sentinel service discovery.
        Tests connection with ping() before returning.

        Args:
            config: Sentinel configuration containing all connection parameters.

        Returns:
            redis.Redis: Connection to the Redis master via sentinel.

        Raises:
            ConnectionException: If sentinel initialization or connection test fails.
        """
        from redis.sentinel import Sentinel

        structlogger.info(
            "redis_connection_factory.sentinel_mode",
            event_info=f"Initializing Redis Sentinel with {len(config.endpoints)} "
            f"sentinel endpoints and service: {config.sentinel_service}",
        )

        # Configuration for Sentinel connection
        connection_kwargs: Dict[str, Any] = {
            "socket_timeout": config.socket_timeout,
        }

        sentinel_kwargs: Optional[Dict] = None
        if config.iam_credentials_provider is not None:
            structlogger.debug("redis_connection_factory.sentinel_iam_auth_enabled")
            sentinel_kwargs = {"credential_provider": config.iam_credentials_provider}
        else:
            connection_kwargs.update(
                {
                    "username": config.username,
                    "password": config.password,
                }
            )

        # SSL configuration
        if config.use_ssl:
            connection_kwargs.update(
                {
                    "ssl": config.use_ssl,
                    "ssl_certfile": config.ssl_certfile,
                    "ssl_keyfile": config.ssl_keyfile,
                    "ssl_ca_certs": config.ssl_ca_certs,
                }
            )

        # Configuration for Redis client (master/replica)
        client_kwargs = {
            "db": config.db,
            "decode_responses": config.decode_responses,
            "socket_timeout": config.socket_timeout,
        }

        # Create Sentinel instance
        try:
            sentinel = Sentinel(
                config.endpoints, sentinel_kwargs=sentinel_kwargs, **connection_kwargs
            )
            master = sentinel.master_for(config.sentinel_service, **client_kwargs)

            # Test the connection
            master.ping()

        except Exception as e:
            raise ConnectionException(f"Error initializing Redis Sentinel: {e}")

        return master

    @classmethod
    def _create_standard_connection(
        cls,
        config: StandardRedisConfig,
    ) -> redis.Redis:
        """Create a standard Redis connection.

        Args:
            config: Standard configuration containing all connection parameters.

        Returns:
            redis.Redis: Configured Redis connection.
        """
        structlogger.info(
            "redis_connection_factory.standard_mode",
            event_info="Initializing Redis connection",
        )
        # Build connection arguments
        connection_args: Dict[str, Any] = {
            "host": config.host,
            "port": int(config.port),
            "db": config.db,
            "socket_timeout": float(config.socket_timeout),
            "ssl": config.use_ssl,
            "ssl_certfile": config.ssl_certfile,
            "ssl_keyfile": config.ssl_keyfile,
            "ssl_ca_certs": config.ssl_ca_certs,
            "decode_responses": config.decode_responses,
        }

        if config.iam_credentials_provider is not None:
            structlogger.debug("redis_connection_factory.standard_iam_auth_enabled")
            connection_args.update(
                {"credential_provider": config.iam_credentials_provider}
            )
        else:
            connection_args.update(
                {
                    "password": config.password,
                    "username": config.username,
                }
            )

        try:
            standard_redis = redis.StrictRedis(**connection_args)
        except Exception as e:
            raise ConnectionException(f"Error initializing Redis connection: {e}")

        return standard_redis

    @classmethod
    def _log_cluster_db_warning(cls, deployment_mode: DeploymentMode, db: int) -> None:
        """Log warning if db parameter is set in cluster mode."""
        if deployment_mode == DeploymentMode.CLUSTER and db != 0:
            structlogger.warning(
                "redis_connection_factory.cluster_db_ignored",
                event_info=f"Database parameter 'db={db}' ignored in cluster mode. "
                "Redis Cluster only supports database 0.",
            )



import asyncio
import time
from pathlib import Path
from typing import TYPE_CHECKING, Any, Dict, List, Optional, Union

import structlog
from tqdm import tqdm

from rasa.core.channels import CollectingOutputChannel, UserMessage
from rasa.core.config.available_endpoints import AvailableEndpoints
from rasa.core.config.configuration import Configuration
from rasa.core.exceptions import AgentNotReady
from rasa.core.persistor import StorageType
from rasa.dialogue_understanding.commands import Command
from rasa.dialogue_understanding.utils import set_record_commands_and_prompts
from rasa.dialogue_understanding_test.du_test_case import (
    DialogueUnderstandingOutput,
    DialogueUnderstandingTestCase,
    DialogueUnderstandingTestStep,
)
from rasa.dialogue_understanding_test.du_test_result import (
    DialogueUnderstandingTestResult,
)
from rasa.dialogue_understanding_test.test_case_simulation.test_case_tracker_simulator import (  # noqa: E501
    TestCaseTrackerSimulator,
    TestCaseTrackerSimulatorException,
    TestCaseTrackerSimulatorResult,
)
from rasa.dialogue_understanding_test.utils import filter_metadata
from rasa.e2e_test.e2e_test_case import (
    KEY_STUB_CUSTOM_ACTIONS,
    Metadata,
)
from rasa.e2e_test.e2e_test_runner import E2ETestRunner
from rasa.e2e_test.utils.fixture_utils import get_fixtures_for_test_case
from rasa.shared.agents.agent_setup import AgentsConnectionCleanup
from rasa.shared.core.events import UserUttered
from rasa.shared.core.trackers import DialogueStateTracker
from rasa.shared.nlu.constants import PREDICTED_COMMANDS, PROMPTS
from rasa.shared.utils.llm import create_tracker_for_user_step
from rasa.utils.endpoints import EndpointConfig

if TYPE_CHECKING:
    from rasa.core.agent import Agent
    from rasa.e2e_test.e2e_test_case import TestCaseFixtures

structlogger = structlog.get_logger()


class DialogueUnderstandingTestRunner:
    """Dialogue Understanding test suite runner.

    Responsible for executing dialogue understanding test cases by simulating
    conversations and validating expected outputs against actual responses.

    Attributes:
        agent: The Rasa agent instance used for processing messages
    """

    def __init__(
        self,
        model_path: Optional[str] = None,
        model_server: Optional[EndpointConfig] = None,
        remote_storage: Optional[StorageType] = None,
        endpoints: Optional[AvailableEndpoints] = None,
        sub_agents_path: Optional[Union[Path, str]] = None,
    ) -> None:
        """Initializes the Dialogue Understanding test suite runner.

        Args:
            model_path: Path to the model.
            model_server: Model server configuration.
            remote_storage: Remote storage to use for model retrieval.
            endpoints: Endpoints configuration.
        """
        import rasa.core.agent

        self._check_action_server(endpoints)
        sub_agents = Configuration.get_instance().available_agents

        async def load_agent_with_agents_connection_cleanup() -> "Agent":
            # Create a wrapper coroutine that handles graceful agents connection
            # cleanup, this prevents abrupt closure when asyncio.run finishes.
            async with AgentsConnectionCleanup():
                return await rasa.core.agent.load_agent(
                    model_path=model_path,
                    model_server=model_server,
                    remote_storage=remote_storage,
                    endpoints=endpoints,
                    sub_agents=sub_agents,
                )
            # Defensive return for mypy - never actually reached
            return

        self.agent = asyncio.run(load_agent_with_agents_connection_cleanup())
        if not self.agent.is_ready():
            raise AgentNotReady(
                "Agent needs to be prepared before usage. "
                "Please check that the agent was able to "
                "load the trained model."
            )

    @staticmethod
    def _check_action_server(endpoints: AvailableEndpoints) -> None:
        """Check if the action server is reachable."""
        are_custom_actions_stubbed = (
            endpoints
            and endpoints.action
            and endpoints.action.kwargs.get(KEY_STUB_CUSTOM_ACTIONS)
        )
        if endpoints and not are_custom_actions_stubbed:
            E2ETestRunner._action_server_is_reachable(
                endpoints, "dialogue_understanding_test_runner"
            )

    async def run_test_cases(
        self,
        test_cases: List[DialogueUnderstandingTestCase],
        fixtures_per_test: List["TestCaseFixtures"],
        metadata: List[Metadata],
    ) -> List[DialogueUnderstandingTestResult]:
        """Run the dialogue understanding tests.

        Args:
            test_cases: List of test cases.
            fixtures_per_test: Per-file resolved fixtures (conftest + local).
            metadata: List of metadata.

        Returns:
            List[DialogueUnderstandingTestResult]: List of test results.
        """
        results = []
        output_channel = CollectingOutputChannel()

        structlogger.info("Starting dialogue understanding tests.")

        for i in tqdm(range(len(test_cases))):
            test_case = test_cases[i]

            # Find the fixtures for the test case
            fixtures = get_fixtures_for_test_case(test_case, fixtures_per_test)

            # set up the tracker by simulating the conversation
            tracker_simulator = TestCaseTrackerSimulator(self.agent, test_case)
            await tracker_simulator.initialize_tracker(fixtures)
            try:
                simulation_result = await tracker_simulator.simulate_test_case(metadata)
            except TestCaseTrackerSimulatorException as e:
                structlogger.error(
                    "Failed to simulate test case. Skipping test case.",
                    error=str(e),
                )
                continue

            # run the actual test case
            test_result = await self.run_test_case(
                test_case, metadata, simulation_result, output_channel
            )
            results.append(test_result)

        structlogger.info("Finished dialogue understanding tests.")

        return results

    async def run_test_case(
        self,
        test_case: DialogueUnderstandingTestCase,
        metadata: List[Metadata],
        simulation_result: TestCaseTrackerSimulatorResult,
        output_channel: CollectingOutputChannel,
    ) -> DialogueUnderstandingTestResult:
        """Runs dialogue understanding test case.

        Args:
            test_case: The test case to run.
            metadata: List of metadata.
            simulation_result: The test case tracker.
            output_channel: The output channel.

        Returns:
            A dialogue understanding test result.
        """
        test_passed = True

        sender_id = simulation_result.sender_id
        user_uttered_event_indices = simulation_result.user_uttered_event_indices
        test_case_tracker = await self.agent.tracker_store.retrieve(sender_id)

        for user_step_index, user_step in enumerate(
            test_case.iterate_over_user_steps()
        ):
            if user_step_index >= len(user_uttered_event_indices):
                # the conversation could not be simulated completely and was
                # stopped early, no further user messages can be evaluated
                return DialogueUnderstandingTestResult(
                    test_case=test_case, passed=test_passed
                )

            # create and save the tracker at the time just
            # before the user message was sent
            step_sender_id = f"{sender_id}_{user_step_index}"
            await create_tracker_for_user_step(
                step_sender_id,
                self.agent,
                test_case_tracker,
                user_uttered_event_indices[user_step_index],
            )

            # Total latency of a message roundtrip
            latency = None

            # send the user message
            try:
                start = time.time()
                await self._send_user_message(
                    step_sender_id,
                    test_case,
                    user_step,
                    metadata,
                    output_channel=output_channel,
                )
                end = time.time()
                latency = end - start
            except Exception as e:
                structlogger.error(
                    "dialogue_understanding_test_runner.send_user_message.failed",
                    test_case=test_case.full_name(),
                    user_message=user_step.text,
                    error=str(e),
                )
                # as sending the user message failed, we cannot continue with the test
                # as subsequent user message do not match the actual conversation
                # return a test result up until this point
                return DialogueUnderstandingTestResult(
                    test_case=test_case, passed=test_passed
                )

            # get the dialogue understanding output
            tracker = await self.agent.tracker_store.retrieve(step_sender_id)
            dialogue_understanding_output = self.get_dialogue_understanding_output(
                tracker, user_uttered_event_indices[user_step_index], latency
            )
            user_step.dialogue_understanding_output = dialogue_understanding_output

            # check if we have a command match
            if not user_step.has_passed():
                test_passed = False

        return DialogueUnderstandingTestResult(test_case=test_case, passed=test_passed)

    def get_dialogue_understanding_output(
        self,
        tracker: DialogueStateTracker,
        index_user_uttered_event: int,
        latency: Optional[float] = None,
    ) -> Optional[DialogueUnderstandingOutput]:
        """Returns the dialogue understanding output.

        Creates the dialogue understanding output from the commands and prompts
        added to the user uttered event.

        Args:
            tracker: The tracker for the current test step.
            index_user_uttered_event: The index of the user uttered event.

        Returns:
        The dialogue understanding output with commands and optionally prompts.
        """
        user_uttered_event = self._get_user_uttered_event_from_tracker(
            tracker, index_user_uttered_event
        )
        if user_uttered_event is None:
            return None

        predicted_commands: Dict[str, Any] = user_uttered_event.parse_data.get(  # type:ignore[assignment]
            PREDICTED_COMMANDS, {}
        )
        # convert the predicted commands to Command objects
        commands = {}
        for component, list_of_commands in predicted_commands.items():
            # remove any duplicate commands
            commands[component] = list(
                set(
                    [Command.command_from_json(command) for command in list_of_commands]
                )
            )

        return DialogueUnderstandingOutput(
            commands=commands,
            prompts=user_uttered_event.parse_data.get(PROMPTS, []),
            latency=latency,
        )

    @staticmethod
    def _get_user_uttered_event_from_tracker(
        tracker: DialogueStateTracker,
        index_user_uttered_event: int,
    ) -> Optional[UserUttered]:
        """Returns the user uttered event from the tracker at the given index."""
        if (
            tracker.events is None
            or len(list(tracker.events)) <= index_user_uttered_event
        ):
            return None

        user_uttered_event = list(tracker.events)[index_user_uttered_event]

        if not user_uttered_event or not isinstance(user_uttered_event, UserUttered):
            return None

        return user_uttered_event

    async def _send_user_message(
        self,
        sender_id: str,
        test_case: DialogueUnderstandingTestCase,
        user_step: DialogueUnderstandingTestStep,
        metadata: List[Metadata],
        output_channel: CollectingOutputChannel = CollectingOutputChannel(),
    ) -> None:
        """Sends a user message to the agent."""
        metadata_for_step = filter_metadata(test_case, user_step, metadata, sender_id)

        user_message = UserMessage(
            user_step.text,
            output_channel,
            sender_id,
            metadata=metadata_for_step,
        )

        with set_record_commands_and_prompts():
            await self.agent.handle_message(user_message)



import typing
from contextlib import contextmanager
from typing import Any, Dict, Generator, List, Optional, Text

import structlog

from rasa.dialogue_understanding.constants import (
    RASA_RECORD_COMMANDS_AND_PROMPTS_ENV_VAR_NAME,
)
from rasa.shared.constants import ROUTE_TO_CALM_SLOT
from rasa.shared.core.trackers import DialogueStateTracker
from rasa.shared.nlu.constants import (
    COMMANDS,
    KEY_COMPONENT_NAME,
    KEY_LLM_RESPONSE_METADATA,
    KEY_PROMPT_NAME,
    KEY_SYSTEM_PROMPT,
    KEY_USER_PROMPT,
    PREDICTED_COMMANDS,
    PROMPTS,
)
from rasa.shared.nlu.training_data.message import Message
from rasa.shared.providers.llm.llm_response import LLMResponse
from rasa.utils.common import get_bool_env_variable

if typing.TYPE_CHECKING:
    from rasa.dialogue_understanding.commands import Command

record_commands_and_prompts = get_bool_env_variable(
    RASA_RECORD_COMMANDS_AND_PROMPTS_ENV_VAR_NAME, False
)

structlogger = structlog.get_logger()

DEFAULT_MAX_CLARIFICATION_OPTIONS = 3
MAX_CLARIFICATION_OPTIONS_SLOT_NAME = "max_clarification_options"


@contextmanager
def set_record_commands_and_prompts() -> Generator:
    global record_commands_and_prompts
    record_commands_and_prompts = True
    try:
        yield
    finally:
        record_commands_and_prompts = False


def add_commands_to_message_parse_data(
    message: Message, component_name: str, commands: List["Command"]
) -> None:
    """Add commands to the message parse data.

    Commands are only added in case the flag 'record_commands_and_prompts' is set.
    Example of predicted commands in the message parse data:
        Message(data={
          PREDICTED_COMMANDS: {
            "MultiStepLLMCommandGenerator": [
                {"command": "set_slot", "name": "slot_name", "value": "slot_value"},
            ],
            "NLUCommandAdapter": [
                {"command": "start_flow", "name": "test_flow"},
            ]
          }
        })
    """
    # only set commands if the flag "record_commands_and_prompts" is set to True
    if not record_commands_and_prompts:
        return

    commands_as_dict = [command.as_dict() for command in commands]

    if message.get(PREDICTED_COMMANDS) is not None:
        predicted_commands = message.get(PREDICTED_COMMANDS)
        if component_name in predicted_commands:
            predicted_commands[component_name].extend(commands_as_dict)
        else:
            predicted_commands[component_name] = commands_as_dict
    else:
        predicted_commands = {component_name: commands_as_dict}

    message.set(
        PREDICTED_COMMANDS,
        predicted_commands,
        add_to_output=True,
    )


def add_prompt_to_message_parse_data(
    message: Message,
    component_name: str,
    prompt_name: str,
    user_prompt: str,
    system_prompt: Optional[str] = None,
    llm_response: Optional[LLMResponse] = None,
) -> None:
    """Add prompt to the message parse data.

    Prompt is only added in case the flag 'record_commands_and_prompts' is set.
    Example of prompts in the message parse data:
        Message(data={
            PROMPTS: [
                {
                    "component_name": "MultiStepLLMCommandGenerator",
                    "prompt_name": "fill_slots_prompt",
                    "user_prompt": "...",
                    "system_prompt": "...",
                    "llm_response_metadata": { ... }
                },
                {
                    "component_name": "MultiStepLLMCommandGenerator",
                    "prompt_name": "handle_flows_prompt",
                    "user_prompt": "...",
                    "system_prompt": "...",
                    "llm_response_metadata": { ... }
                },
                {
                    "component_name": "SingleStepLLMCommandGenerator",
                    "prompt_name": "prompt_template",
                    "user_prompt": "...",
                    "system_prompt": "...",
                    "llm_response_metadata": { ... }
                }
            ]
        }
    )
    """
    # Only set prompt if the flag "record_commands_and_prompts" is set to True.
    if not record_commands_and_prompts:
        return

    # Construct the dictionary with prompt details.
    prompt_data: Dict[Text, Any] = {
        KEY_COMPONENT_NAME: component_name,
        KEY_PROMPT_NAME: prompt_name,
        KEY_USER_PROMPT: user_prompt,
        KEY_LLM_RESPONSE_METADATA: llm_response.to_dict() if llm_response else None,
        **({KEY_SYSTEM_PROMPT: system_prompt} if system_prompt else {}),
    }

    # Get or create a top-level "prompts" list.
    prompts = message.get(PROMPTS) or []
    prompts.append(prompt_data)

    # Update the message with the new prompts list.
    message.set(PROMPTS, prompts, add_to_output=True)


def _handle_via_nlu_in_coexistence(
    tracker: Optional[DialogueStateTracker], message: Message
) -> bool:
    """Check if the message should be handled by the NLU subsystem in coexistence mode."""  # noqa: E501
    from rasa.dialogue_understanding.commands import (
        NoopCommand,
        SetSlotCommand,
    )

    if not tracker:
        return False

    commands = message.get(COMMANDS, [])

    # If coexistence routing slot is not active, this setup doesn't
    # support dual routing -> default to CALM
    if not tracker.has_coexistence_routing_slot:
        structlogger.debug(
            "utils.handle_via_nlu_in_coexistence"
            ".tracker_missing_route_session_to_calm_slot",
            event_info=(
                f"Tracker doesn't have the '{ROUTE_TO_CALM_SLOT}' slot.Routing to CALM."
            ),
            route_session_to_calm=commands,
        )
        return False

    # Check if the routing decision is stored in the tracker slot
    # If slot is true -> route to CALM
    # If slot is false -> route to DM1
    value = tracker.get_slot(ROUTE_TO_CALM_SLOT)
    if value is not None:
        structlogger.debug(
            "utils.handle_via_nlu_in_coexistence"
            ".tracker_route_session_to_calm_slot_value",
            event_info=(
                f"Tracker slot '{ROUTE_TO_CALM_SLOT}' set to '{value}'. "
                f"Routing to "
                f"{'CALM' if value else 'NLU system'}."
            ),
            route_session_to_calm_value_in_tracker=value,
        )
        return not value

    # Non-sticky routing to DM1 is only allowed if NoopCommand is the sole predicted
    # command. In that case, route to DM1
    if len(commands) == 1 and commands[0].get("command") == NoopCommand.command():
        structlogger.debug(
            "utils.handle_via_nlu_in_coexistence.noop_command_detected",
            event_info="NoopCommand found. Routing to NLU system non-sticky.",
            commands=commands,
        )
        return True

    # If the slot was reset (e.g. new session), try to infer routing from
    # attached commands. Look for a SetSlotCommand targeting the ROUTE_TO_CALM_SLOT
    for command in message.get(COMMANDS, []):
        # If slot is true -> route to CALM
        # If slot is false -> route to DM1
        if (
            command.get("command") == SetSlotCommand.command()
            and command.get("name") == ROUTE_TO_CALM_SLOT
        ):
            structlogger.debug(
                "utils.handle_via_nlu_in_coexistence.set_slot_command_detected",
                event_info=(
                    f"SetSlotCommand setting the '{ROUTE_TO_CALM_SLOT}' to "
                    f"'{command.get('value')}'. "
                    f"Routing to "
                    f"{'CALM' if command.get('value') else 'NLU system'}."
                ),
                commands=commands,
            )
            return not command.get("value")

    # If no routing info is available -> default to CALM
    structlogger.debug(
        "utils.handle_via_nlu_in_coexistence.no_routing_info_available",
        event_info="No routing info available. Routing to CALM.",
        commands=commands,
    )
    return False


def assemble_options_string(names: List[str], conjunction: str = "and") -> str:
    """Concatenate options to a human-readable string."""
    option_message = ""
    for i, name in enumerate(names):
        if i == 0:
            option_message += name
        elif i == len(names) - 1:
            option_message += f" {conjunction} {name}"
        else:
            option_message += f", {name}"
    return option_message


def limit_clarification_options(
    tracker: "DialogueStateTracker", names: List[str]
) -> List[str]:
    """Limit the number of clarification options based on the slot value.

    Args:
        tracker: The dialogue state tracker.
        names: The list of clarification option names.

    Returns:
        The limited list of names, or the original list if no valid limit is set.
    """
    max_options_slot = tracker.get_slot(MAX_CLARIFICATION_OPTIONS_SLOT_NAME)
    try:
        max_options = int(max_options_slot)
    except (ValueError, TypeError):
        structlogger.debug(
            "utils.limit_clarification_options.invalid_slot_value",
            slot_value=max_options_slot,
            event_info=(
                f"Slot '{MAX_CLARIFICATION_OPTIONS_SLOT_NAME}' has invalid value. "
                f"Falling back to default '{DEFAULT_MAX_CLARIFICATION_OPTIONS}'."
            ),
        )
        max_options = DEFAULT_MAX_CLARIFICATION_OPTIONS

    return names[:max_options]



import logging
from pathlib import Path
from typing import Any, Dict, Optional, Text

from rasa.utils.installation_utils import check_for_installation_issues
from rasa.utils.tensorflow import TENSORFLOW_AVAILABLE

if TENSORFLOW_AVAILABLE:
    check_for_installation_issues()
    import tensorflow as tf
    from tqdm import tqdm
else:
    # Placeholder values when TensorFlow is not available
    tf = None
    tqdm = None

import rasa.shared.utils.io

logger = logging.getLogger(__name__)


if TENSORFLOW_AVAILABLE:

    class RasaTrainingLogger(tf.keras.callbacks.Callback):
        """Callback for logging the status of training."""

        def __init__(self, epochs: int, silent: bool) -> None:
            """Initializes the callback.

            Args:
                epochs: Total number of epochs.
                silent: If 'True' the entire progressbar wrapper is disabled.
            """
            super().__init__()

            disable = silent or rasa.shared.utils.io.is_logging_disabled()
            self.progress_bar = tqdm(range(epochs), desc="Epochs", disable=disable)

        def on_epoch_end(
            self, epoch: int, logs: Optional[Dict[Text, Any]] = None
        ) -> None:
            """Updates the logging output on every epoch end.

            Args:
                epoch: The current epoch.
                logs: The training metrics.
            """
            self.progress_bar.update(1)
            self.progress_bar.set_postfix(logs)

        def on_train_end(self, logs: Optional[Dict[Text, Any]] = None) -> None:
            """Closes the progress bar after training.

            Args:
                logs: The training metrics.
            """
            self.progress_bar.close()

    class RasaModelCheckpoint(tf.keras.callbacks.Callback):
        """Callback for saving intermediate model checkpoints."""

        def __init__(self, checkpoint_dir: Path) -> None:
            """Initializes the callback.

            Args:
                checkpoint_dir: Directory to store checkpoints to.
            """
            super().__init__()

            self.checkpoint_file = checkpoint_dir / "checkpoint.weights.h5"
            self.best_metrics_so_far: Dict[Text, Any] = {}

        def on_epoch_end(
            self, epoch: int, logs: Optional[Dict[Text, Any]] = None
        ) -> None:
            """Save the model on epoch end if the model has improved.

            Args:
                epoch: The current epoch.
                logs: The training metrics.
            """
            if self._does_model_improve(logs):
                logger.debug(f"Creating model checkpoint at epoch={epoch + 1} ...")
                # Ensure model is built before saving weights
                if not self.model.built:
                    # Build the model with dummy data to ensure it's built
                    import tensorflow as tf

                    dummy_input = tf.zeros((1, 1))
                    _ = self.model(dummy_input)

                # Ensure the directory exists before saving
                import os

                os.makedirs(os.path.dirname(self.checkpoint_file), exist_ok=True)
                self.model.save_weights(self.checkpoint_file, overwrite=True)

        def _does_model_improve(self, curr_results: Dict[Text, Any]) -> bool:
            """Checks whether the current results are better than the best so far.

            Results are considered better if each metric is
            equal or better than the best so far, and at least one is better.

            Args:
                curr_results: The training metrics for this epoch.
            """
            curr_metric_names = [
                k
                for k in curr_results.keys()
                if k.startswith("val") and (k.endswith("_acc") or k.endswith("_f1"))
            ]
            # the "val" prefix is prepended to metrics in fit
            # if _should_eval returns true
            # for this particular epoch
            if len(curr_metric_names) == 0:
                # the metrics are not validation metrics
                return False
            # initialize best_metrics_so_far with the first results
            if not self.best_metrics_so_far:
                for metric_name in curr_metric_names:
                    self.best_metrics_so_far[metric_name] = float(
                        curr_results[metric_name]
                    )
                    return True

            at_least_one_improved = False
            improved_metrics = {}
            for metric_name in self.best_metrics_so_far.keys():
                if (
                    float(curr_results[metric_name])
                    < self.best_metrics_so_far[metric_name]
                ):
                    # at least one of the values is worse
                    return False
                if (
                    float(curr_results[metric_name])
                    > self.best_metrics_so_far[metric_name]
                ):
                    at_least_one_improved = True
                    improved_metrics[metric_name] = float(curr_results[metric_name])

            # all current values >= previous best and at least one is better
            if at_least_one_improved:
                self.best_metrics_so_far.update(improved_metrics)
            return at_least_one_improved
else:
    # Placeholder classes when TensorFlow is not available
    RasaTrainingLogger = None  # type: ignore
    RasaModelCheckpoint = None  # type: ignore



import json
from typing import Any, Dict, List, Optional, Text

import jsonschema
from ruamel.yaml.error import (
    MarkedYAMLError,
    MarkedYAMLFutureWarning,
    MarkedYAMLWarning,
)


class RasaException(Exception):
    """Base exception class for all errors raised by Rasa Pro.

    These exceptions result from invalid use cases and will be reported
    to the users, but will be ignored in telemetry.
    """

    def __init__(self, message: str, suppress_stack_trace: bool = False, **kwargs: Any):
        """Initialize the exception.

        Args:
            message: The error message.
            suppress_stack_trace: If True, the stack trace will be suppressed in logs.
            **kwargs: Additional keyword arguments (e.g., cause for exception chaining).
        """
        Exception.__init__(self, message)
        self.suppress_stack_trace = suppress_stack_trace


class RasaCoreException(RasaException):
    """Basic exception for errors raised by Rasa Core."""


class InvalidParameterException(RasaException, ValueError):
    """Raised when an invalid parameter is used."""


class YamlException(RasaException):
    """Raised if there is an error reading yaml."""

    def __init__(self, filename: Optional[Text] = None) -> None:
        """Create exception.

        Args:
        filename: optional file the error occurred in
        """
        self.filename = filename

    def file_error_message(self) -> str:
        if self.filename:
            return f"Error in '{self.filename}'."
        else:
            return "Error found."

    def __str__(self) -> str:
        msg = self.file_error_message()
        if self.__cause__:
            msg += f" {self.__cause__}"
        return msg


class YamlSyntaxException(YamlException):
    """Raised when a YAML file can not be parsed properly due to a syntax error."""

    def __init__(
        self,
        filename: Optional[Text] = None,
        underlying_yaml_exception: Optional[Exception] = None,
    ) -> None:
        super(YamlSyntaxException, self).__init__(filename)

        self.underlying_yaml_exception = underlying_yaml_exception

    def __str__(self) -> Text:
        exception_text = self.file_error_message()
        exception_text += " Failed to read YAML."

        if self.underlying_yaml_exception:
            if isinstance(
                self.underlying_yaml_exception,
                (MarkedYAMLError, MarkedYAMLWarning, MarkedYAMLFutureWarning),
            ):
                self.underlying_yaml_exception.note = None
            if isinstance(
                self.underlying_yaml_exception,
                (MarkedYAMLWarning, MarkedYAMLFutureWarning),
            ):
                self.underlying_yaml_exception.warn = None
            exception_text += f" {self.underlying_yaml_exception}"

        if self.filename:
            exception_text = exception_text.replace(
                'in "<unicode string>"', f'in "{self.filename}"'
            )

        exception_text += (
            "\n\nYou can use https://yamlchecker.com/ to validate the "
            "YAML syntax of your file."
        )
        return exception_text


class FileNotFoundException(RasaException, FileNotFoundError):
    """Raised when a file, expected to exist, doesn't exist."""


class FileIOException(RasaException):
    """Raised if there is an error while doing file IO."""


class InvalidPromptTemplateException(RasaException):
    """Raised when a custom prompt template file cannot be read."""

    def __init__(self, file_path: str, resolved_path: str) -> None:
        """Initialize the exception.

        Args:
            file_path: The path to the prompt template file that was specified.
            resolved_path: The resolved absolute path to the file.
        """
        message = (
            f"Failed to read custom prompt template from `{file_path}`. "
            f"Resolved path: `{resolved_path}`. "
            f"Please ensure the file exists and is readable."
        )
        super().__init__(message)
        self.file_path = file_path
        self.resolved_path = resolved_path


class InvalidConfigException(ValueError, RasaException):
    """Raised if an invalid configuration is encountered."""


class UnsupportedFeatureException(RasaCoreException):
    """Raised if a requested feature is not supported."""


class SchemaValidationError(RasaException, jsonschema.ValidationError):
    """Raised if schema validation via `jsonschema` failed."""


class InvalidEntityFormatException(RasaException, json.JSONDecodeError):
    """Raised if the format of an entity is invalid."""

    def __init__(self, msg: str, doc: str = "", pos: int = 0):
        """Initialize the exception.

        Args:
            msg: The error message.
            doc: The document that caused the error.
            pos: The position in the document where the error occurred.
        """
        RasaException.__init__(self, msg)
        json.JSONDecodeError.__init__(self, msg, doc, pos)

    @classmethod
    def create_from(
        cls, other: json.JSONDecodeError, msg: Text
    ) -> "InvalidEntityFormatException":
        """Creates `InvalidEntityFormatException` from `JSONDecodeError`."""
        return cls(msg, other.doc, other.pos)


class ConnectionException(RasaException):
    """Raised when a connection to a 3rd party service fails.

    It's used by our broker and tracker store classes, when
    they can't connect to services like postgres, dynamoDB, mongo.
    """


class ProviderClientAPIException(RasaException):
    """For errors during API interactions with LLM / embedding providers.

    Attributes:
        original_exception (Exception): The original exception that was
            caught and led to this custom exception.
        message: (Optional[str]): Optional explanation of the error.
    """

    def __init__(
        self,
        original_exception: Exception,
        message: Optional[str] = None,
        info: Optional[Dict[Text, Any]] = None,
    ):
        super().__init__(
            f"{message if message is not None else ''}"
            f"\nOriginal error: {original_exception})"
        )
        self.message = message
        self.original_exception = original_exception
        self.info = info

    def __str__(self) -> Text:
        s = f"{self.__class__.__name__}:"
        if self.message is not None:
            s += f"\n{self.message}"
        s += f"\nOriginal error: {self.original_exception}\n"
        if self.info:
            s += f"\nInfo: \n{self.info}\n"
        return s


class LLMToolResponseDecodeError(ProviderClientAPIException):
    """Raised when a JSON decoding error occurs in LLM tool response."""


class ProviderClientValidationError(RasaException):
    """Raised for errors that occur during validation of the API client."""


class FinetuningDataPreparationException(RasaException):
    """Raised when there is an error in data preparation for fine-tuning."""


class AgentInitializationException(RasaException):
    """Raised when there is an error during the initialization of an agent."""


class AgentAuthInitializationException(RasaException):
    """Raised when there is an error during the initialization of agent auth client."""


class AuthenticationError(RasaException):
    """Raised when there is an authentication error."""


class DuplicateFixtureException(RasaException):
    """Raised when duplicate fixture names are found."""

    def __init__(self, fixture_names: List[str], source_file: str) -> None:
        """Initialize the exception.

        Args:
            fixture_names: The name(s) of the duplicate fixture(s).
            source_file: The source file of the duplicate fixture(s).
        """
        if len(fixture_names) == 1:
            message = (
                f"Duplicate fixture '{fixture_names[0]}' found in '{source_file}'."
            )
        else:
            names = ", ".join(f"'{name}'" for name in fixture_names)
            message = f"Duplicate fixtures {names} found in '{source_file}'."

        super().__init__(message)



"""Validation functions for agent configurations with reduced redundancies."""

import os
import urllib.parse
from collections import Counter
from typing import Any, Dict, List, NoReturn, Set

import jinja2.exceptions
import structlog
from pydantic import ValidationError as PydanticValidationError

from rasa.agents.exceptions import (
    AgentNameFlowConflictException,
    DuplicatedAgentNameException,
)
from rasa.core.available_agents import (
    DEFAULT_AGENTS_CONFIG_FOLDER,
    AgentConfig,
    AgentConfiguration,
    AgentConnections,
    AgentInfo,
    AvailableAgents,
    ProtocolConfig,
)
from rasa.core.config.available_endpoints import AvailableEndpoints
from rasa.core.config.configuration import Configuration
from rasa.exceptions import ValidationError
from rasa.shared.agents.auth.utils import validate_secrets_in_params
from rasa.shared.utils.llm import get_prompt_template, validate_jinja2_template
from rasa.shared.utils.yaml import read_config_file

# Initialize logger
structlogger = structlog.get_logger()

# Centralized allowed keys configuration to eliminate duplication
ALLOWED_KEYS = {
    "agent": {"name", "protocol", "description"},
    "configuration": {
        "llm",
        "prompt_template",
        "module",
        "timeout",
        "max_retries",
        "agent_card",
        "auth",
        "include_date_time",
        "timezone",
    },
    "connections": {"mcp_servers"},
}


def validate_agent_names_unique(agents: List[AgentConfig]) -> None:
    """Validate that agent names are unique across all loaded agents.

    Args:
        agents: List of agent configurations.

    Raises:
        DuplicatedAgentNameException: If agent names are not unique.
    """
    agent_names = [agent_config.agent.name for agent_config in agents]
    name_counts = Counter(agent_names)

    duplicated_names = [name for name, count in name_counts.items() if count > 1]

    if duplicated_names:
        raise DuplicatedAgentNameException(duplicated_names)


def validate_agent_names_not_conflicting_with_flows(
    agents: Dict[str, AgentConfig], flow_names: Set[str]
) -> None:
    """Validate that agent names do not conflict with flow names.

    Args:
        agents: Dictionary of agent configurations.
        flow_names: Set of flow names to check against.

    Raises:
        AgentNameFlowConflictException: If agent names conflict with flow names.
    """
    conflicting_names = [
        agent_name for agent_name in agents.keys() if agent_name in flow_names
    ]

    if conflicting_names:
        raise AgentNameFlowConflictException(conflicting_names)


def _validate_mcp_config(agent_config: AgentConfig) -> None:
    """Validate MCP-specific configuration requirements."""
    agent_name = agent_config.agent.name

    # Check connections.mcp_servers exists
    if agent_config.connections is None or agent_config.connections.mcp_servers is None:
        raise ValidationError(
            code="agent.validation.mcp.missing_connections",
            event_info=f"For protocol 'RASA', agent '{agent_name}' must have "
            "'connections.mcp_servers' configured.",
        )

    # Check mcp_servers list is not empty
    if not agent_config.connections.mcp_servers:
        raise ValidationError(
            code="agent.validation.mcp.empty_servers_list",
            event_info=f"For protocol 'RASA', agent '{agent_name}' must have "
            "at least one MCP server configured in 'connections.mcp_servers'.",
        )

    # Validate each MCP server configuration
    for i, server in enumerate(agent_config.connections.mcp_servers):
        if not server.name:
            raise ValidationError(
                code="agent.validation.mcp.server_missing_name",
                event_info=f"For protocol 'RASA', agent '{agent_name}' MCP server "
                f"at index {i} must have a 'name' field.",
            )


def _validate_a2a_config(agent_config: AgentConfig) -> None:
    """Validate A2A-specific configuration requirements."""
    agent_name = agent_config.agent.name

    # Check configuration.agent_card exists
    if (
        agent_config.configuration is None
        or agent_config.configuration.agent_card is None
    ):
        raise ValidationError(
            code="agent.validation.a2a.missing_agent_card",
            event_info=f"For protocol 'A2A', agent '{agent_name}' must have "
            "'configuration.agent_card' specified.",
        )

    # Validate agent_card path or URL
    agent_card = agent_config.configuration.agent_card
    if not agent_card:
        raise ValidationError(
            code="agent.validation.a2a.empty_agent_card",
            event_info=f"Agent '{agent_name}' has empty 'agent_card' value",
        )

    # Check if it's a URL
    parsed_url = urllib.parse.urlparse(agent_card)
    if parsed_url.scheme and parsed_url.netloc:
        # It's a URL, validate format
        if parsed_url.scheme not in ["http", "https"]:
            raise ValidationError(
                code="agent.validation.a2a.invalid_agent_card_url",
                event_info=f"Agent '{agent_name}' has invalid URL scheme "
                f"'{parsed_url.scheme}' for 'agent_card'",
            )
    else:
        # It's a file path, check if it exists
        if not os.path.exists(agent_card):
            raise ValidationError(
                code="agent.validation.a2a.agent_card_file_not_found",
                event_info=f"Agent '{agent_name}' has 'agent_card' file that doesn't "
                f"exist: {agent_card}",
            )

    # Validate right use of secrets in auth configuration
    if agent_config.configuration.auth:
        validate_secrets_in_params(
            agent_config.configuration.auth, f"agent '{agent_name}'"
        )


def _validate_prompt_template_syntax(prompt_path: str, agent_name: str) -> None:
    """Validate Jinja2 syntax of a prompt template file."""
    try:
        # Use a simple default template, as we're assuming
        # that the default templates are valid
        default_template = "{{ content }}"
        template_content = get_prompt_template(
            prompt_path,
            default_template,
            log_source_component=f"agent.validation.{agent_name}",
            log_source_method="init",
        )
        validate_jinja2_template(template_content)

    except jinja2.exceptions.TemplateSyntaxError as e:
        raise ValidationError(
            code="agent.validation.prompt_template_syntax_error",
            event_info=(
                f"Agent '{agent_name}' has invalid Jinja2 template syntax at line "
                f"{e.lineno}: {e.message}"
            ),
        ) from e
    except Exception as e:
        raise ValidationError(
            code="agent.validation.optional.prompt_template_error",
            event_info=(f"Agent '{agent_name}' has error reading prompt template: {e}"),
        ) from e


def _validate_optional_keys(agent_config: AgentConfig) -> None:
    """Validate optional keys in agent configuration."""
    agent_name = agent_config.agent.name

    # Validate prompt_template if present
    if agent_config.configuration and agent_config.configuration.prompt_template:
        prompt_path = agent_config.configuration.prompt_template
        if not os.path.exists(prompt_path):
            # If reading the custom prompt fails,
            # allow fallback to default prompt template
            structlogger.warning(
                "agent.validation.optional.prompt_template_file_not_found",
                agent_name=agent_name,
                prompt_path=prompt_path,
                event_info=(
                    f"Prompt template file not found: {prompt_path}. "
                    f"Agent will use default template."
                ),
            )
            # Don't raise ValidationError, allow fallback to default template
            return

        # Validate Jinja2 syntax
        _validate_prompt_template_syntax(prompt_path, agent_name)

    # Validate module if present
    if agent_config.configuration and agent_config.configuration.module:
        import importlib

        module_name = agent_config.configuration.module
        try:
            module_path, class_name = module_name.rsplit(".", 1)
            getattr(importlib.import_module(module_path), class_name)
        except (ImportError, AttributeError) as e:
            raise ValidationError(
                code="agent.validation.optional.module_not_found",
                event_info=f"Agent '{agent_name}' has module '{module_name}' "
                f"that could not be imported: {e}",
            )


def _validate_llm_references(
    llm_config: Dict[str, Any],
    endpoints: "AvailableEndpoints",
    agent_name: str,
) -> None:
    """Validate LLM configuration references against endpoints."""
    if "model_group" in llm_config:
        from rasa.engine.validation import (
            _validate_component_model_client_config_has_references_to_endpoints,
        )

        component_config = {"llm": llm_config}
        _validate_component_model_client_config_has_references_to_endpoints(
            component_config, "llm", component_name=agent_name
        )


def _validate_mcp_server_references(
    mcp_servers: list, endpoints: "AvailableEndpoints", agent_name: str
) -> None:
    """Validate MCP server references against endpoints."""
    if not endpoints.mcp_servers:
        raise ValidationError(
            code="agent.validation.endpoints.no_mcp_servers",
            event_info=(
                f"Agent '{agent_name}' references MCP servers but no MCP "
                "servers are defined in endpoints.yml."
            ),
        )

    available_mcp_server_names = [server.name for server in endpoints.mcp_servers]

    for i, mcp_server in enumerate(mcp_servers):
        server_name = mcp_server.name
        if server_name not in available_mcp_server_names:
            raise ValidationError(
                code="agent.validation.endpoints.invalid_mcp_server",
                event_info=(
                    f"MCP server '{server_name}' at index {i} for Agent "
                    f"'{agent_name}' does not exist in endpoints.yml. Available MCP "
                    f"servers: {', '.join(available_mcp_server_names)}"
                ),
            )


def _handle_pydantic_validation_error(
    error: PydanticValidationError, agent_name: str
) -> NoReturn:
    """Handle specific Pydantic validation errors that are actually possible.

    Args:
        error: The Pydantic validation error to handle
        agent_name: Name of the agent for error messages
    """
    missing_fields = []
    invalid_protocol = False
    type_error = None

    for pydantic_error in error.errors():
        error_type = pydantic_error["type"]
        field_path = ".".join(str(loc) for loc in pydantic_error["loc"])

        if error_type == "missing":
            for field in ["name", "protocol", "description"]:
                if field in field_path:
                    missing_fields.append(field)
        elif error_type == "enum" and "protocol" in field_path:
            invalid_protocol = True
        elif error_type in ["string_type", "int_parsing"]:
            type_error = (field_path, pydantic_error["msg"])

    # Handle missing required fields
    if missing_fields:
        raise ValidationError(
            code="agent.validation.mandatory.fields_missing",
            event_info=(
                f"Agent '{agent_name}' is missing required fields "
                f"in agent section: {', '.join(missing_fields)}"
            ),
        )

    # Handle invalid protocol
    elif invalid_protocol:
        raise ValidationError(
            code="agent.validation.pydantic.invalid_protocol",
            event_info=(
                f"Agent '{agent_name}' has invalid protocol value. "
                "Supported protocols: MCP, A2A"
            ),
        )

    # Handle type errors
    elif type_error:
        field, msg = type_error
        raise ValidationError(
            code="agent.validation.pydantic.type_error",
            event_info=(
                f"Agent '{agent_name}' has invalid type for field " f"'{field}': {msg}"
            ),
        )

    # Handle other Pydantic validation errors
    else:
        raise ValidationError(
            code="agent.validation.pydantic.failed",
            event_info=f"Agent '{agent_name}' validation failed: {error}",
        )


def _validate_endpoint_references(agent_config: AgentConfig) -> None:
    """Validate that LLM and MCP server references in agent config are valid."""
    agent_name = agent_config.agent.name
    endpoints = Configuration.get_instance().endpoints
    if not endpoints.config_file_path:
        # If no endpoints were loaded (e.g., `data validate` without --endpoints), skip
        # endpoint reference checks
        return

    # Validate LLM configuration references
    if agent_config.configuration and agent_config.configuration.llm:
        _validate_llm_references(agent_config.configuration.llm, endpoints, agent_name)

    # Validate MCP server references
    if agent_config.connections and agent_config.connections.mcp_servers:
        _validate_mcp_server_references(
            agent_config.connections.mcp_servers, endpoints, agent_name
        )


def _validate_section_keys(
    data: Dict[str, Any], section: str, allowed_keys: set
) -> None:
    """Generic function to validate keys in a specific section."""
    if section not in data:
        return

    section_data = data[section]
    if not isinstance(section_data, dict):
        return

    additional_keys = set(section_data.keys()) - allowed_keys
    if additional_keys:
        agent_name = data.get("agent", {}).get("name", "unknown")
        raise ValidationError(
            code=f"agent.validation.structure.additional_{section}_keys",
            event_info=(
                f"Agent '{agent_name}' contains additional keys in "
                f"'{section}' section: {', '.join(sorted(additional_keys))}"
            ),
        )


def _validate_mandatory_fields(data: Dict[str, Any], agent_name: str) -> None:
    """Validate that all mandatory fields are present in the agent section."""
    if "agent" not in data:
        raise ValidationError(
            code="agent.validation.mandatory.agent_section_missing",
            event_info=f"Agent '{agent_name}' is missing 'agent' section",
        )

    agent_data = data["agent"]
    if not isinstance(agent_data, dict):
        raise ValidationError(
            code="agent.validation.mandatory.agent_section_invalid",
            event_info=(
                f"Agent '{agent_name}' has invalid 'agent' section - "
                "must be a dictionary"
            ),
        )

    # Check for required fields (protocol is optional due to default in model)
    missing_fields = []
    for field in ["name", "description"]:
        if field not in agent_data or not agent_data[field]:
            missing_fields.append(field)

    if missing_fields:
        raise ValidationError(
            code="agent.validation.mandatory.fields_missing",
            event_info=(
                f"Agent '{agent_name}' is missing required fields in agent section: "
                f"{', '.join(missing_fields)}"
            ),
        )


def _validate_no_additional_keys_raw_data(data: Dict[str, Any]) -> None:
    """Validate that no additional, unexpected keys are present in the raw data."""
    # Use the generic validation function for each section
    _validate_section_keys(data, "agent", ALLOWED_KEYS["agent"])
    _validate_section_keys(data, "configuration", ALLOWED_KEYS["configuration"])
    _validate_section_keys(data, "connections", ALLOWED_KEYS["connections"])


def _validate_folder_structure(agent_folder: str) -> None:
    """Validate agent folder structure."""
    if not os.path.exists(agent_folder):
        raise ValidationError(
            code="agent.validation.folder.not_found",
            event_info=f"Agent folder does not exist: {agent_folder}",
        )

    if not os.path.isdir(agent_folder):
        raise ValidationError(
            code="agent.validation.folder.not_directory",
            event_info=f"Agent folder is not a directory: {agent_folder}",
        )


def _validate_config_file_exists(config_path: str, agent_name: str) -> None:
    """Validate config.yml file exists."""
    if not os.path.isfile(config_path):
        raise ValidationError(
            code="agent.validation.folder.missing_config",
            event_info=f"Agent '{agent_name}' is missing 'config.yml' file",
        )


def validate_agent_config(agent_config: AgentConfig) -> None:
    """Validate an agent configuration using all applicable validators."""
    protocol = agent_config.agent.protocol

    # Run protocol-specific validation
    if protocol == ProtocolConfig.RASA:
        _validate_mcp_config(agent_config)
    elif protocol == ProtocolConfig.A2A:
        _validate_a2a_config(agent_config)

    # Run optional keys validation
    _validate_optional_keys(agent_config)

    # Run endpoint references validation
    _validate_endpoint_references(agent_config)


def validate_agent_folder(agent_folder: str = DEFAULT_AGENTS_CONFIG_FOLDER) -> None:
    """Validate all agent configurations in a folder."""
    # Validate folder structure
    _validate_folder_structure(agent_folder)

    # Scan for agent folders
    for agent_folder_name in os.listdir(agent_folder):
        agent_path = os.path.join(agent_folder, agent_folder_name)

        if not os.path.isdir(agent_path):
            continue

        config_path = os.path.join(agent_path, "config.yml")

        # Validate config file exists
        _validate_config_file_exists(config_path, agent_folder_name)

        # Read and validate the config content
        try:
            # First read the raw YAML data to validate structure
            data = read_config_file(config_path)

            # Validate no additional keys
            _validate_no_additional_keys_raw_data(data)

            # Validate mandatory fields before creating Pydantic models
            _validate_mandatory_fields(data, agent_folder_name)

            agent_config = AvailableAgents.from_dict(data)

            # Validate the agent config (protocol-specific and endpoint references)
            validate_agent_config(agent_config)
        except PydanticValidationError as e:
            _handle_pydantic_validation_error(e, agent_folder_name)
        except Exception as e:
            # Handle non-Pydantic exceptions
            raise ValidationError(
                code="agent.validation.folder.config_validation_failed",
                event_info=f"Agent '{agent_folder_name}' validation failed: {e}",
            )


def validate_agent_config_data(data: Dict[str, Any]) -> "AgentConfig":
    """Validate agent configuration data."""
    try:
        # Validate no additional keys before creating Pydantic models
        _validate_no_additional_keys_raw_data(data)

        # Create the agent config (this will trigger Pydantic validation)
        agent_config = AgentConfig(
            agent=AgentInfo(**data.get("agent", {})),
            configuration=AgentConfiguration(**data.get("configuration", {}))
            if data.get("configuration")
            else None,
            connections=AgentConnections(**data.get("connections", {}))
            if data.get("connections")
            else None,
        )

        return agent_config

    except PydanticValidationError as e:
        _handle_pydantic_validation_error(e, "Agent configuration")
    except Exception:
        # Re-raise other exceptions
        raise



import base64
import os
from dataclasses import dataclass
from typing import AsyncIterator, Dict, Optional

import aiohttp
import structlog
from aiohttp import ClientTimeout

from rasa.core.channels.voice_stream.audio_bytes import HERTZ, RasaAudioBytes
from rasa.core.channels.voice_stream.tts.tts_engine import (
    TTSEngine,
    TTSEngineConfig,
    TTSError,
)
from rasa.shared.constants import CARTESIA_API_KEY_ENV_VAR
from rasa.shared.exceptions import ConnectionException

structlogger = structlog.get_logger()


@dataclass
class CartesiaTTSConfig(TTSEngineConfig):
    model_id: Optional[str] = None
    version: Optional[str] = None
    endpoint: Optional[str] = None


class CartesiaTTS(TTSEngine[CartesiaTTSConfig]):
    session: Optional[aiohttp.ClientSession] = None
    required_env_vars = (CARTESIA_API_KEY_ENV_VAR,)
    ws: Optional[aiohttp.ClientWebSocketResponse] = None
    streaming_input: bool = True

    def __init__(self, config: Optional[CartesiaTTSConfig] = None):
        super().__init__(config)
        timeout = ClientTimeout(total=self.config.timeout)
        # Have to create this class-shared session lazily at run time otherwise
        # the async event loop doesn't work
        if self.__class__.session is None or self.__class__.session.closed:
            self.__class__.session = aiohttp.ClientSession(timeout=timeout)

    @staticmethod
    def get_request_headers(config: CartesiaTTSConfig) -> dict[str, str]:
        """Get headers for WebSocket connection."""
        cartesia_api_key = os.environ[CARTESIA_API_KEY_ENV_VAR]
        return {
            "Cartesia-Version": str(config.version),
            "X-API-Key": str(cartesia_api_key),
        }

    async def connect(self, config: Optional[CartesiaTTSConfig] = None) -> None:
        """Establish WebSocket connection to Cartesia TTS."""
        headers = self.get_request_headers(self.config)
        ws_url = self.config.endpoint

        if self.session is None:
            raise ConnectionException("Client session is not initialized")

        if not ws_url:
            raise ConnectionException("Cartesia endpoint not configured")

        self.ws = await self.session.ws_connect(
            ws_url,
            headers=headers,
            timeout=float(self.config.timeout) if self.config.timeout else 30,
        )

    async def close_connection(self) -> None:
        """Close WebSocket connection if it exists."""
        if self.ws and not self.ws.closed:
            await self.ws.close()
            self.ws = None

    async def _send_tts_request(self, text: str, flush: bool) -> None:
        """Send TTS request to Cartesia via WebSocket."""
        if not self.ws or self.ws.closed:
            raise TTSError("WebSocket connection not established")

        message: Dict[str, object] = {
            "model_id": self.config.model_id,
            "voice": {
                "mode": "id",
                "id": self.config.voice,
            },
            "language": self.config.language,
            "output_format": {
                "container": "raw",
                "encoding": "pcm_mulaw",
                "sample_rate": HERTZ,
            },
            "context_id": "rasa-voice-stream",
        }

        if flush:
            message["flush"] = True
        else:
            message["transcript"] = text
            message["continue"] = True

        await self.ws.send_json(message)

    async def send_text_chunk(self, text: str) -> None:
        """Send text to TTS engine for continuous streaming.

        This sends text to Cartesia but doesn't return anything.
        Audio will be available via stream_audio().
        """
        await self._send_tts_request(text, flush=False)

    async def signal_text_done(self) -> None:
        """Signal TTS engine to process any remaining buffered text.

        This tells Cartesia that all text has been sent and to finish processing.
        """
        await self._send_tts_request("", flush=True)

    async def stream_audio(self) -> AsyncIterator[RasaAudioBytes]:
        """Stream audio output from the TTS engine.

        This continuously yields audio chunks as they arrive from Cartesia.
        Stops when it receives a done message or flush completion.
        """
        if not self.ws or self.ws.closed:
            raise TTSError("WebSocket connection not established")

        try:
            async for msg in self.ws:
                data = msg.json()
                msg_type = data.get("type")

                if msg_type == "chunk":
                    # Audio data chunk - decode base64 and yield
                    base64_audio = data.get("data")
                    if base64_audio:
                        audio_bytes = base64.b64decode(base64_audio)
                        yield self.engine_bytes_to_rasa_audio_bytes(audio_bytes)

                elif msg_type == "done":
                    # TTS processing is done
                    structlogger.debug("cartesia.stream_audio.done")
                    return

                elif msg_type == "error":
                    # Error occurred
                    error_msg = data.get("error", "Unknown error")
                    structlogger.error("cartesia.stream_audio.error", error=error_msg)
                    raise TTSError(f"Cartesia TTS error: {error_msg}")

        except Exception as e:
            structlogger.error("cartesia.stream_audio.error", error=str(e))
            raise TTSError(f"Error during audio streaming: {e}")

    async def synthesize(
        self, text: str, config: Optional[CartesiaTTSConfig] = None
    ) -> AsyncIterator[RasaAudioBytes]:
        """Generate speech from text using a remote TTS system."""
        if not self.ws or self.ws.closed:
            raise TTSError("WebSocket connection not established")

        await self.send_text_chunk(text)
        await self.signal_text_done()

        async for audio_chunk in self.stream_audio():
            yield audio_chunk

    def engine_bytes_to_rasa_audio_bytes(self, chunk: bytes) -> RasaAudioBytes:
        """Convert the generated tts audio bytes into rasa audio bytes."""
        return RasaAudioBytes(chunk)

    @staticmethod
    def get_default_config() -> CartesiaTTSConfig:
        return CartesiaTTSConfig(
            language="en",
            voice="f786b574-daa5-4673-aa0c-cbe3e8534c02",
            timeout=30,
            model_id="sonic-3",
            version="2025-04-16",
            endpoint="wss://api.cartesia.ai/tts/websocket",
        )

    @classmethod
    def from_config_dict(cls, config: Dict) -> "CartesiaTTS":
        return cls(CartesiaTTSConfig.from_dict(config))



from __future__ import annotations

import dataclasses
from typing import Any, Dict, Text

import rasa.shared.utils.io
from rasa.engine.graph import ExecutionContext, GraphComponent
from rasa.engine.storage.resource import Resource
from rasa.engine.storage.storage import ModelStorage
from rasa.shared.core.domain import Domain


@dataclasses.dataclass
class Forms:
    """Holds the forms of the domain."""

    data: Dict[Text, Any]

    def fingerprint(self) -> Text:
        """Returns a fingerprint of the responses."""
        return rasa.shared.utils.io.get_dictionary_fingerprint(self.data)

    def get(self, key: Text, default: Any) -> Any:
        """Returns the value for the given key."""
        return self.data.get(key, default)


class FormsProvider(GraphComponent):
    """Provides forms during training and inference time."""

    @classmethod
    def create(
        cls,
        config: Dict[Text, Any],
        model_storage: ModelStorage,
        resource: Resource,
        execution_context: ExecutionContext,
    ) -> FormsProvider:
        """Creates component (see parent class for full docstring)."""
        return cls()

    def provide(self, domain: Domain) -> Forms:
        """Returns the forms from the given domain."""
        return Forms(data=domain.forms)



"""This module regroups custom validation functions, and it is
loaded as an extension of the pykwalify library:

https://pykwalify.readthedocs.io/en/latest/extensions.html#extensions
"""

from typing import Any, Dict, List, Optional, Union

from pykwalify.errors import SchemaError

BOT_UTTERED_KEY = "bot_uttered"
BUTTONS_KEY = "buttons"
FLOW_STARTED_KEY = "flow_started"
PATTERN_CLARIFICATION_CONTAINS_KEY = "pattern_clarification_contains"


def run_common_dict_assertion_validations(
    assertion_dict: Dict[str, Any], assertion_key: str
) -> Optional[SchemaError]:
    if not assertion_dict:
        return SchemaError(
            f"The '{assertion_key}' assertion cannot be an empty dictionary."
        )

    operator = assertion_dict.get("operator")
    if operator is None:
        return SchemaError(
            f"The 'operator' key is missing in the '{assertion_key}' assertion."
        )
    if operator.lower() not in {"all", "any"}:
        return SchemaError(
            f"The 'operator' in the '{assertion_key}' assertion "
            f"must be either 'all' or 'any' (case insensitive)."
        )

    flow_ids = assertion_dict.get("flow_ids")

    if flow_ids is None:
        return SchemaError(
            f"The 'flow_ids' key is missing in the '{assertion_key}' assertion."
        )

    if not isinstance(flow_ids, list):
        return SchemaError(
            f"The 'flow_ids' in the '{assertion_key}' assertion must be a list."
        )

    if len(flow_ids) == 0:
        return SchemaError(
            f"The 'flow_ids' in the '{assertion_key}' assertion must "
            f"be a non-empty list."
        )

    return None


def require_assertion_keys(
    assertions: List[Dict[str, Any]], _: Dict, __: str
) -> Union[SchemaError, bool]:
    """Validates that certain assertion keys are not mapped to empty values."""
    for assertion in assertions:
        if not isinstance(assertion, dict):
            # this is handled by other validation rules
            continue

        bot_uttered_dict = assertion.get(BOT_UTTERED_KEY)
        if BOT_UTTERED_KEY in assertion and isinstance(bot_uttered_dict, dict):
            if not bot_uttered_dict:
                return SchemaError(
                    f"The '{BOT_UTTERED_KEY}' assertion is an empty dictionary."
                )

            if BUTTONS_KEY in bot_uttered_dict and not bot_uttered_dict.get(
                BUTTONS_KEY
            ):
                return SchemaError(
                    f"The '{BUTTONS_KEY}' key in the '{BOT_UTTERED_KEY}' assertion "
                    f"is mapped to a null value or empty list."
                )

        flow_started_assertion = assertion.get(FLOW_STARTED_KEY)
        if flow_started_assertion is not None:
            if not isinstance(flow_started_assertion, str) and not isinstance(
                flow_started_assertion, dict
            ):
                return SchemaError(
                    f"The '{FLOW_STARTED_KEY}' assertion must be either a string "
                    f"or a dictionary."
                )

            if (
                isinstance(flow_started_assertion, str)
                and not flow_started_assertion.strip()
            ):
                return SchemaError(
                    f"The '{FLOW_STARTED_KEY}' assertion cannot be an empty string."
                )

            if isinstance(flow_started_assertion, dict):
                result = run_common_dict_assertion_validations(
                    flow_started_assertion, FLOW_STARTED_KEY
                )
                if result is not None:
                    return result

                operator = flow_started_assertion.get("operator")
                flow_ids = flow_started_assertion.get("flow_ids", [])
                if operator and operator.lower() == "all" and len(flow_ids) != 1:
                    return SchemaError(
                        f"When using the 'all' operator in the '{FLOW_STARTED_KEY}' "
                        f"assertion, exactly one flow ID must be provided."
                    )

        pattern_clarification_contains = assertion.get(
            PATTERN_CLARIFICATION_CONTAINS_KEY
        )
        if pattern_clarification_contains is not None:
            if not isinstance(pattern_clarification_contains, list) and not isinstance(
                pattern_clarification_contains, dict
            ):
                return SchemaError(
                    f"The '{PATTERN_CLARIFICATION_CONTAINS_KEY}' assertion must be "
                    f"either a list or a dictionary."
                )

            if isinstance(pattern_clarification_contains, list):
                if not pattern_clarification_contains:
                    return SchemaError(
                        f"The '{PATTERN_CLARIFICATION_CONTAINS_KEY}' assertion "
                        f"cannot be an empty list."
                    )

                if not all(
                    isinstance(item, str) and item.strip()
                    for item in pattern_clarification_contains
                ):
                    return SchemaError(
                        f"All items in the '{PATTERN_CLARIFICATION_CONTAINS_KEY}' "
                        f"list must be non-empty strings."
                    )

            if isinstance(pattern_clarification_contains, dict):
                result = run_common_dict_assertion_validations(
                    pattern_clarification_contains, PATTERN_CLARIFICATION_CONTAINS_KEY
                )
                if result is not None:
                    return result

    return True



from __future__ import annotations

import os
import re
from typing import Any, Dict, List, Optional, Text, Union

import structlog

from rasa.shared.constants import SECRET_DATA_FORMAT_PATTERN
from rasa.shared.utils.io import resolve_environment_variables
from rasa.shared.utils.yaml import read_config_file
from rasa.tracing.constants import (
    ENDPOINTS_TRACING_KEY,
    LANGFUSE_CONFIG_BASE_URL_KEY,
    LANGFUSE_CONFIG_DEBUG_KEY,
    LANGFUSE_CONFIG_ENVIRONMENT_KEY,
    LANGFUSE_CONFIG_MEDIA_UPLOAD_THREAD_COUNT_KEY,
    LANGFUSE_CONFIG_PRIVATE_KEY,
    LANGFUSE_CONFIG_PUBLIC_KEY,
    LANGFUSE_CONFIG_RELEASE_KEY,
    LANGFUSE_CONFIG_SAMPLE_RATE_KEY,
    LANGFUSE_CONFIG_TIMEOUT_KEY,
    LANGFUSE_ENV_VAR_DEBUG,
    LANGFUSE_ENV_VAR_MEDIA_UPLOAD_THREAD_COUNT,
    LANGFUSE_ENV_VAR_OTEL_HOST,
    LANGFUSE_ENV_VAR_PUBLIC_KEY,
    LANGFUSE_ENV_VAR_RELEASE,
    LANGFUSE_ENV_VAR_SAMPLE_RATE,
    LANGFUSE_ENV_VAR_SECRET_KEY,
    LANGFUSE_ENV_VAR_TIMEOUT,
    LANGFUSE_ENV_VAR_TRACING_ENVIRONMENT,
    TRACING_TYPE_LANGFUSE,
)
from rasa.tracing.exceptions import (
    DuplicateTracingConfigException,
    InvalidLangfuseConfigException,
)
from rasa.utils.endpoints import EndpointConfig

structlogger = structlog.get_logger()

# https://langfuse.com/docs/observability/features/environments#data-model
LANGFUSE_VALID_ENVIRONMENT_PATTERN = re.compile(r"^(?!langfuse)[a-z0-9-_]+$")


def configure_langfuse(
    endpoints_file: str, langfuse_environment_name: Optional[str] = None
) -> None:
    """Configure Langfuse callback for litellm from endpoints file.

    Args:
        endpoints_file: Path to endpoints configuration file.
        langfuse_environment_name: Optional environment name to force for Langfuse
            tracing (e.g. "E2E test", "DU test"). If provided, this overrides
            any environment configured in the endpoints file.
    """
    langfuse_config = _get_langfuse_config(endpoints_file)

    if not langfuse_config:
        structlogger.debug(
            "langfuse_configuration.langfuse_config_not_found",
            event_info=(
                f"No Langfuse configuration found in {endpoints_file}, "
                f"Langfuse will not be configured."
            ),
        )
        return

    if not _is_langfuse_available():
        return

    config_values = _extract_langfuse_config_values(langfuse_config)

    _validate_langfuse_config(config_values)

    public_key = config_values[LANGFUSE_CONFIG_PUBLIC_KEY]
    private_key = config_values[LANGFUSE_CONFIG_PRIVATE_KEY]

    resolved_keys = _resolve_environment_variables(public_key, private_key)
    _set_langfuse_environment_variables(
        config_values,
        resolved_keys,
        langfuse_environment_name=langfuse_environment_name,
    )
    _configure_litellm_callback()

    structlogger.info(
        "langfuse_configuration.langfuse_config_configured",
        event_info=(f"Langfuse has been configured from {endpoints_file}."),
    )


def _is_langfuse_available() -> bool:
    """Check if Langfuse is available."""
    try:
        import langfuse  # noqa: F401
    except ImportError as e:
        structlogger.error(
            "langfuse_configuration.langfuse_not_available",
            event_info=(
                "Langfuse is not available. "
                "Please install langfuse via `pip install rasa-pro[monitoring]`. "
                "Langfuse will not be configured."
            ),
            error=str(e),
        )
        return False
    return True


def _get_langfuse_config(endpoints_file: str) -> Optional[EndpointConfig]:
    """Read Langfuse configuration from endpoints file."""
    if not endpoints_file:
        return None

    try:
        content = read_config_file(endpoints_file)
    except FileNotFoundError:
        structlogger.debug(
            "langfuse_configuration.endpoints_file_not_found",
            event_info=(
                f"Endpoints file not found: {endpoints_file}, "
                f"Langfuse will not be configured."
            ),
        )
        return None
    except Exception:
        structlogger.error(
            "langfuse_configuration.endpoints_file_read_error",
            event_info=(
                f"Error reading endpoints file: {endpoints_file}, "
                f"Langfuse will not be configured."
            ),
        )
        return None

    if content.get(ENDPOINTS_TRACING_KEY) is None:
        return None

    tracing_config: Union[List[Dict[Text, Any]], Dict[Text, Any]] = content.get(
        ENDPOINTS_TRACING_KEY, []
    )

    tracing_configs = _parse_tracing_configs(tracing_config)

    langfuse_configs = [
        config for config in tracing_configs if config.type == TRACING_TYPE_LANGFUSE
    ]

    if len(langfuse_configs) == 0:
        return None

    if len(langfuse_configs) > 1:
        _log_multiple_langfuse_configs_error(endpoints_file)
        raise DuplicateTracingConfigException(
            f"Multiple Langfuse configs found in {endpoints_file}, "
            "which is not supported. Only one Langfuse config is allowed."
        )

    return langfuse_configs[0]


def _parse_tracing_configs(
    tracing_config: Union[List[Dict[Text, Any]], Dict[Text, Any]],
) -> List[EndpointConfig]:
    """Parse tracing configuration into list of EndpointConfig objects."""
    if isinstance(tracing_config, list):
        return [EndpointConfig.from_dict(item) for item in tracing_config]
    else:
        return [EndpointConfig.from_dict(tracing_config)]


def _log_multiple_langfuse_configs_error(endpoints_file: str) -> None:
    """Log error when multiple Langfuse configs are found."""
    structlogger.error(
        "langfuse_configuration.multiple_langfuse_configs",
        event_info=(
            f"Multiple Langfuse configs found in {endpoints_file}, "
            f"which is not supported. Only one Langfuse config is allowed."
        ),
    )


def _extract_langfuse_config_values(
    langfuse_config: EndpointConfig,
) -> Dict[str, Any]:
    """Extract all configuration values from Langfuse config."""
    return {
        LANGFUSE_CONFIG_PUBLIC_KEY: langfuse_config.kwargs.get(
            LANGFUSE_CONFIG_PUBLIC_KEY
        ),
        LANGFUSE_CONFIG_PRIVATE_KEY: langfuse_config.kwargs.get(
            LANGFUSE_CONFIG_PRIVATE_KEY
        ),
        LANGFUSE_CONFIG_BASE_URL_KEY: langfuse_config.kwargs.get(
            LANGFUSE_CONFIG_BASE_URL_KEY
        ),
        LANGFUSE_CONFIG_TIMEOUT_KEY: langfuse_config.kwargs.get(
            LANGFUSE_CONFIG_TIMEOUT_KEY
        ),
        LANGFUSE_CONFIG_DEBUG_KEY: langfuse_config.kwargs.get(
            LANGFUSE_CONFIG_DEBUG_KEY
        ),
        LANGFUSE_CONFIG_ENVIRONMENT_KEY: langfuse_config.kwargs.get(
            LANGFUSE_CONFIG_ENVIRONMENT_KEY
        ),
        LANGFUSE_CONFIG_RELEASE_KEY: langfuse_config.kwargs.get(
            LANGFUSE_CONFIG_RELEASE_KEY
        ),
        LANGFUSE_CONFIG_MEDIA_UPLOAD_THREAD_COUNT_KEY: langfuse_config.kwargs.get(
            LANGFUSE_CONFIG_MEDIA_UPLOAD_THREAD_COUNT_KEY
        ),
        LANGFUSE_CONFIG_SAMPLE_RATE_KEY: langfuse_config.kwargs.get(
            LANGFUSE_CONFIG_SAMPLE_RATE_KEY
        ),
    }


def _resolve_environment_variables(
    public_key: str, private_key: str
) -> Dict[str, Union[str, List[Any], Dict[str, Any]]]:
    """Resolve environment variables in public and secret keys."""
    return {
        LANGFUSE_CONFIG_PUBLIC_KEY: resolve_environment_variables(public_key),
        LANGFUSE_CONFIG_PRIVATE_KEY: resolve_environment_variables(private_key),
    }


def _set_langfuse_environment_variables(
    config_values: Dict[str, Any],
    resolved_keys: Dict[str, Optional[str]],
    langfuse_environment_name: Optional[str] = None,
) -> None:
    """Set Langfuse environment variables for LiteLLM integration.

    Logs warnings when overwriting existing environment variables.
    """
    # Map of config keys to (source dict, environment variable name)
    env_var_mappings = [
        (LANGFUSE_CONFIG_PUBLIC_KEY, resolved_keys, LANGFUSE_ENV_VAR_PUBLIC_KEY),
        (LANGFUSE_CONFIG_PRIVATE_KEY, resolved_keys, LANGFUSE_ENV_VAR_SECRET_KEY),
        (LANGFUSE_CONFIG_BASE_URL_KEY, config_values, LANGFUSE_ENV_VAR_OTEL_HOST),
        (LANGFUSE_CONFIG_TIMEOUT_KEY, config_values, LANGFUSE_ENV_VAR_TIMEOUT),
        (LANGFUSE_CONFIG_DEBUG_KEY, config_values, LANGFUSE_ENV_VAR_DEBUG),
        (LANGFUSE_CONFIG_RELEASE_KEY, config_values, LANGFUSE_ENV_VAR_RELEASE),
        (
            LANGFUSE_CONFIG_MEDIA_UPLOAD_THREAD_COUNT_KEY,
            config_values,
            LANGFUSE_ENV_VAR_MEDIA_UPLOAD_THREAD_COUNT,
        ),
        (LANGFUSE_CONFIG_SAMPLE_RATE_KEY, config_values, LANGFUSE_ENV_VAR_SAMPLE_RATE),
    ]

    def _set_env_var(env_var_name: str, value: Optional[str]) -> None:
        if value is None:
            return
        existing_value = os.getenv(env_var_name)
        # Only warn when the value is actually changing to reduce log noise
        if existing_value is not None and existing_value != value:
            structlogger.warning(
                "langfuse_configuration.overwriting_env_var",
                event_info=(
                    f"Overwriting existing environment variable '{env_var_name}' "
                    f"with value from endpoints configuration."
                ),
                env_var_name=env_var_name,
            )
        os.environ[env_var_name] = value

    for config_key, source_dict, env_var_name in env_var_mappings:
        _set_env_var(env_var_name, source_dict.get(config_key))

    # Decide langfuse environment name value once, then set
    environment = (
        langfuse_environment_name
        if langfuse_environment_name is not None
        else config_values.get(LANGFUSE_CONFIG_ENVIRONMENT_KEY)
    )
    _set_env_var(LANGFUSE_ENV_VAR_TRACING_ENVIRONMENT, environment)


def _configure_litellm_callback() -> None:
    """Configure LiteLLM to use Langfuse as a callback."""
    import litellm

    # LiteLLM will automatically trace LLM calls to langfuse if these environment
    # variables are set and the callback is configured
    litellm.success_callback = ["langfuse_otel"]


def _validate_langfuse_config(config_values: Dict[str, Any]) -> None:
    """Validate Langfuse configuration."""
    _validate_required_keys(
        config_values,
        [
            LANGFUSE_CONFIG_PUBLIC_KEY,
            LANGFUSE_CONFIG_PRIVATE_KEY,
            LANGFUSE_CONFIG_BASE_URL_KEY,
        ],
    )

    public_key = config_values[LANGFUSE_CONFIG_PUBLIC_KEY]
    private_key = config_values[LANGFUSE_CONFIG_PRIVATE_KEY]
    _validate_key_syntax(public_key, private_key)
    _validate_environment_name(config_values.get(LANGFUSE_CONFIG_ENVIRONMENT_KEY))


def _validate_environment_name(environment: Optional[str]) -> None:
    """Validate the optional Langfuse environment name if provided.

    The environment must match the regex ^(?!langfuse)[a-z0-9-_]+$ and be at most
    40 characters long.
    """
    if not environment:
        return

    if (
        not isinstance(environment, str)
        or len(environment) > 40
        or not LANGFUSE_VALID_ENVIRONMENT_PATTERN.match(environment)
    ):
        error_message = (
            f"Invalid Langfuse environment: '{environment}'. It can only contain "
            f"lowercase letters, numbers, hyphens, and underscores. It cannot start "
            f"with 'langfuse' prefix and must be at most 40 characters long."
        )
        structlogger.error(
            "langfuse_configuration.invalid_environment",
            event_info=error_message,
            environment=environment,
        )
        raise InvalidLangfuseConfigException(error_message)


def _validate_required_keys(
    config_values: Dict[str, Any], required_keys: List[str]
) -> None:
    """Validate that required keys are provided."""
    missing_keys: List[str] = []
    for key in required_keys:
        if not config_values[key]:
            missing_keys.append(key)

    if missing_keys:
        structlogger.error(
            "tracing.langfuse_config.invalid_config",
            event_info=(
                f"Langfuse config is invalid. The following keys are required and "
                f"currently not configured: {missing_keys}."
            ),
        )
        raise InvalidLangfuseConfigException(
            f"Langfuse config is invalid. The following keys are required and "
            f"currently not configured: {missing_keys}."
        )


def _validate_key_syntax(public_key: str, secret_key: str) -> None:
    """Validate that keys use the correct ${VAR} syntax for environment variables."""
    if not re.match(SECRET_DATA_FORMAT_PATTERN, public_key) or not re.match(
        SECRET_DATA_FORMAT_PATTERN, secret_key
    ):
        structlogger.error(
            "langfuse_configuration.langfuse_config_invalid",
            event_info=(
                f"Langfuse config is invalid: {LANGFUSE_CONFIG_PUBLIC_KEY} and "
                f"{LANGFUSE_CONFIG_PRIVATE_KEY} need to be provided via the "
                f"${{syntax}}."
            ),
        )
        raise InvalidLangfuseConfigException(
            f"Langfuse config is invalid: {LANGFUSE_CONFIG_PUBLIC_KEY} and "
            f"{LANGFUSE_CONFIG_PRIVATE_KEY} need to be provided via the ${{syntax}}."
        )



from __future__ import annotations

import asyncio
import base64
import hmac
import json
from typing import Any, Awaitable, Callable, Dict, Optional, Text

import structlog
from sanic import (  # type: ignore[attr-defined]
    Blueprint,
    HTTPResponse,
    Request,
    Websocket,
    response,
)

from rasa.core.channels import UserMessage
from rasa.core.channels.voice_ready.audiocodes import map_call_params
from rasa.core.channels.voice_ready.utils import CallParameters
from rasa.core.channels.voice_stream.audio_bytes import RasaAudioBytes
from rasa.core.channels.voice_stream.call_state import (
    call_state,
)
from rasa.core.channels.voice_stream.tts.tts_engine import TTSEngine
from rasa.core.channels.voice_stream.util import repack_voice_credentials
from rasa.core.channels.voice_stream.voice_channel import (
    ContinueConversationAction,
    DTMFInputAction,
    EndConversationAction,
    NewAudioAction,
    VoiceChannelAction,
    VoiceInputChannel,
    VoiceOutputChannel,
)
from rasa.shared.utils.common import mark_as_beta_feature

logger = structlog.get_logger(__name__)
PREFERRED_AUDIO_FORMAT = "raw/mulaw"


class AudiocodesVoiceOutputChannel(VoiceOutputChannel):
    @classmethod
    def name(cls) -> str:
        return "audiocodes_stream"

    def _ensure_stream_id(self) -> None:
        """Audiocodes requires a stream ID with playStream messages."""
        if "stream_id" not in call_state.channel_data:
            call_state.channel_data["stream_id"] = 0

    def _increment_stream_id(self) -> None:
        self._ensure_stream_id()
        call_state.channel_data["stream_id"] += 1

    def _get_stream_id(self) -> str:
        self._ensure_stream_id()
        return str(call_state.channel_data["stream_id"])

    def rasa_audio_bytes_to_channel_bytes(
        self, rasa_audio_bytes: RasaAudioBytes
    ) -> bytes:
        return base64.b64encode(rasa_audio_bytes)

    def channel_bytes_to_message(self, recipient_id: str, channel_bytes: bytes) -> str:
        media_message = json.dumps(
            {
                "type": "playStream.chunk",
                "streamId": self._get_stream_id(),
                "audioChunk": channel_bytes.decode("utf-8"),
            }
        )
        return media_message

    async def send_start_marker(self, recipient_id: str) -> None:
        """Send playStream.start before first audio chunk."""
        self._increment_stream_id()
        media_message = json.dumps(
            {
                "type": "playStream.start",
                "streamId": self._get_stream_id(),
                "mediaFormat": PREFERRED_AUDIO_FORMAT,
            }
        )
        logger.debug("Sending start marker", stream_id=self._get_stream_id())
        await self.voice_websocket.send(media_message)

        # This should be set when the bot actually starts speaking
        # however, Audiocodes does not have an event to indicate that.
        # This is an approximation, as the bot will be sent the audio chunks next
        # which are played to the user immediately.
        call_state.is_bot_speaking = True
        VoiceInputChannel._cancel_silence_timeout_watcher()

    async def send_intermediate_marker(self, recipient_id: str) -> None:
        """Audiocodes doesn't need intermediate markers, so do nothing."""
        pass

    async def send_end_marker(self, recipient_id: str) -> None:
        """Send playStream.stop after last audio chunk."""
        media_message = json.dumps(
            {
                "type": "playStream.stop",
                "streamId": self._get_stream_id(),
            }
        )
        logger.debug("Sending end marker", stream_id=self._get_stream_id())
        await self.voice_websocket.send(media_message)


class AudiocodesVoiceInputChannel(VoiceInputChannel):
    @classmethod
    def name(cls) -> str:
        return "audiocodes_stream"

    def __init__(
        self,
        server_url: str,
        asr_config: Dict,
        tts_config: Dict,
        interruptions: Optional[Dict[str, int]] = None,
        token: Optional[Text] = None,
    ):
        mark_as_beta_feature("Audiocodes (audiocodes_stream) Channel")
        super().__init__(
            server_url=server_url,
            asr_config=asr_config,
            tts_config=tts_config,
            interruptions=interruptions,
        )
        self.token = token

    @classmethod
    def from_credentials(
        cls,
        credentials: Optional[Dict[str, Any]],
    ) -> AudiocodesVoiceInputChannel:
        cls.validate_basic_credentials(credentials)
        new_creds = repack_voice_credentials(credentials)
        return cls(**new_creds)

    def channel_bytes_to_rasa_audio_bytes(self, input_bytes: bytes) -> RasaAudioBytes:
        return RasaAudioBytes(base64.b64decode(input_bytes))

    async def collect_call_parameters(
        self, channel_websocket: Websocket
    ) -> Optional[CallParameters]:
        async for message in channel_websocket:
            data = json.loads(message)
            if data["type"] == "session.initiate":
                # contains info about mediaformats
                logger.info(
                    "audiocodes_stream.collect_call_parameters.session.initiate",
                    data=data,
                )
                self._send_accepted(channel_websocket, data)
            elif data["type"] == "activities":
                activities = data["activities"]
                for activity in activities:
                    logger.debug(
                        "audiocodes_stream.collect_call_parameters.activity",
                        data=activity,
                    )
                    if activity["name"] == "start":
                        return map_call_params(activity["parameters"])
            elif data["type"] == "connection.validate":
                # not part of call flow; only sent when integration is created
                logger.info(
                    "audiocodes_stream.collect_call_parameters.connection.validate",
                    event_info="received request to validate integration",
                )
                self._send_validated(channel_websocket, data)
            else:
                logger.warning("audiocodes_stream.unknown_message", data=data)
        return None

    def map_input_message(
        self,
        message: Any,
        ws: Websocket,
    ) -> VoiceChannelAction:
        data = json.loads(message)
        if data["type"] == "activities":
            activities = data["activities"]
            for activity in activities:
                if activity["name"] == "start":
                    # handled in collect_call_parameters
                    pass
                elif activity["name"] == "DTMF":
                    return DTMFInputAction(digit=activity["value"])
                elif activity["name"] == "playFinished":
                    logger.debug("audiocodes_stream.playFinished", data=activity)
                    call_state.is_bot_speaking = False
                    if call_state.should_hangup:
                        logger.info("audiocodes_stream.hangup")
                        self._send_hangup(ws, data)
                        # the conversation should continue until
                        # we receive a end message from audiocodes
                else:
                    logger.warning("audiocodes_stream.unknown_activity", data=activity)
        elif data["type"] == "userStream.start":
            logger.debug("audiocodes_stream.userStream.start", data=data)
            self._send_recognition_started(ws, data)
        elif data["type"] == "userStream.chunk":
            audio_bytes = self.channel_bytes_to_rasa_audio_bytes(data["audioChunk"])
            return NewAudioAction(audio_bytes)
        elif data["type"] == "userStream.stop":
            logger.debug("audiocodes_stream.stop_recognition", data=data)
            self._send_recognition_ended(ws, data)
        elif data["type"] == "session.resume":
            logger.debug("audiocodes_stream.resume", data=data)
            self._send_accepted(ws, data)
        elif data["type"] == "session.end":
            logger.debug("audiocodes_stream.end", data=data)
            return EndConversationAction()
        else:
            logger.warning(
                "audiocodes_stream.map_input_message.unknown_message", data=data
            )

        return ContinueConversationAction()

    def _send_accepted(self, ws: Websocket, data: Dict[Text, Any]) -> None:
        supported_formats = data.get("supportedMediaFormats", [])
        preferred_format = PREFERRED_AUDIO_FORMAT

        if preferred_format not in supported_formats:
            logger.warning(
                "audiocodes_stream.format_not_supported",
                supported_formats=supported_formats,
                preferred_format=preferred_format,
            )
            raise

        payload = {
            "type": "session.accepted",
            "mediaFormat": PREFERRED_AUDIO_FORMAT,
        }
        _schedule_async_task(ws.send(json.dumps(payload)))

    def _send_recognition_started(self, ws: Websocket, data: Dict[Text, Any]) -> None:
        payload = {"type": "userStream.started"}
        _schedule_async_task(ws.send(json.dumps(payload)))

    def _send_recognition_ended(self, ws: Websocket, data: Dict[Text, Any]) -> None:
        payload = {"type": "userStream.stopped"}
        _schedule_async_task(ws.send(json.dumps(payload)))

    def _send_hypothesis(self, ws: Websocket, data: Dict[Text, Any]) -> None:
        """
        TODO: The hypothesis message is sent by the bot to provide partial
        recognition results. Using this message is recommended,
        as VAIC relies on it for performing barge-in.
        """
        pass

    def _send_recognition(self, ws: Websocket, data: Dict[Text, Any]) -> None:
        """
        TODO: The recognition message is sent by the bot to provide
        the final recognition result. Using this message is recommended
        mainly for logging purposes.
        """
        pass

    def _send_hangup(self, ws: Websocket, data: Dict[Text, Any]) -> None:
        payload = {
            "conversationId": data["conversationId"],
            "type": "activities",
            "activities": [{"type": "event", "name": "hangup"}],
        }
        _schedule_async_task(ws.send(json.dumps(payload)))

    def _send_validated(self, ws: Websocket, data: Dict[Text, Any]) -> None:
        payload = {
            "type": "connection.validated",
            "success": True,
        }
        _schedule_async_task(ws.send(json.dumps(payload)))

    def create_output_channel(
        self, voice_websocket: Websocket, tts_engine: TTSEngine
    ) -> VoiceOutputChannel:
        return AudiocodesVoiceOutputChannel(
            voice_websocket,
            tts_engine,
            self.tts_cache,
        )

    def _is_token_valid(self, token: Optional[Text]) -> bool:
        # If no token is set, always return True
        if not self.token:
            return True

        # Token is required, but not provided
        if not token:
            return False

        return hmac.compare_digest(str(self.token), str(token))

    def blueprint(
        self, on_new_message: Callable[[UserMessage], Awaitable[Any]]
    ) -> Blueprint:
        """Defines a Sanic blueprint"""
        blueprint = Blueprint("audiocodes_stream", __name__)

        @blueprint.route("/", methods=["GET"])
        async def health(_: Request) -> HTTPResponse:
            return response.json({"status": "ok"})

        @blueprint.websocket("/websocket")  # type: ignore
        async def receive(request: Request, ws: Websocket) -> None:
            if not self._is_token_valid(request.token):
                logger.error(
                    "audiocodes_stream.invalid_token",
                    invalid_token=request.token,
                )
                await ws.close(code=1008, reason="Invalid token")
                return

            logger.info(
                "audiocodes_stream.receive", event_info="Started websocket connection"
            )
            try:
                await self.run_audio_streaming(on_new_message, ws)
            except Exception as e:
                logger.exception(
                    "audiocodes_stream.receive",
                    message="Error during audio streaming",
                    error=e,
                )
                await ws.close(code=1011, reason="Error during audio streaming")
                raise

        return blueprint


def _schedule_async_task(coro: Awaitable[Any]) -> None:
    """Helper function to schedule a coroutine in the event loop.

    Args:
        coro: The coroutine to schedule
    """
    loop = asyncio.get_running_loop()
    loop.call_soon_threadsafe(lambda: loop.create_task(coro))



from __future__ import annotations

import json
from dataclasses import asdict, dataclass
from datetime import datetime, timezone
from pathlib import Path
from typing import Optional

from rasa.constants import RASA_DIR_NAME


@dataclass
class ProjectInfo:
    """Metadata persisted for the builder about the current project.

    - first_used_utc_iso: RFC3339/ISO8601 string of when the project was first used.
    """

    first_used_utc_iso: Optional[str] = None

    @classmethod
    def from_json(cls, data: str) -> "ProjectInfo":
        payload = json.loads(data) if data else {}
        return cls(**payload)

    def to_json(self) -> str:
        return json.dumps(asdict(self))

    def first_used_dt(self) -> Optional[datetime]:
        if not self.first_used_utc_iso:
            return None
        value = self.first_used_utc_iso.strip()
        if value.endswith("Z"):
            value = value[:-1] + "+00:00"
        try:
            dt = datetime.fromisoformat(value)
        except Exception:
            return None
        if dt.tzinfo is None:
            dt = dt.replace(tzinfo=timezone.utc)
        return dt.astimezone(timezone.utc)


def _project_info_path(project_folder: Path) -> Path:
    return project_folder / RASA_DIR_NAME / "project_info.json"


def load_project_info(project_folder: Path) -> ProjectInfo:
    path = _project_info_path(project_folder)
    try:
        if path.exists():
            return ProjectInfo.from_json(path.read_text(encoding="utf-8"))
    except Exception:
        pass
    return ProjectInfo()


def save_project_info(project_folder: Path, info: ProjectInfo) -> None:
    path = _project_info_path(project_folder)
    path.parent.mkdir(parents=True, exist_ok=True)
    path.write_text(info.to_json(), encoding="utf-8")


def ensure_first_used(project_folder: Path) -> datetime:
    info = load_project_info(project_folder)
    if existing := info.first_used_dt():
        return existing

    now = datetime.now(timezone.utc)
    info.first_used_utc_iso = now.isoformat()
    save_project_info(project_folder, info)
    return now



import importlib_resources

from rasa.constants import PACKAGE_NAME

INKEEP_API_KEY_ENV_VAR = "INKEEP_API_KEY"

INKEEP_RAG_RESPONSE_SCHEMA_PATH = str(
    importlib_resources.files(PACKAGE_NAME)
    .joinpath("builder")
    .joinpath("document_retrieval")
    .joinpath("inkeep-rag-response-schema.json")
)


INKEEP_DOCUMENT_RETRIEVAL_MODEL = "inkeep-rag"



from datetime import datetime
from pathlib import Path
from typing import Any, Dict, List, Optional, Tuple

import langfuse
import structlog
import yaml  # type: ignore[import-untyped]
from langfuse import Evaluation
from langfuse._client.datasets import DatasetClient
from langfuse.experiment import (
    ExperimentItem,
    ExperimentItemResult,
    ExperimentResult,
)

from rasa.builder.copilot.models import (
    ResponseCategory,
)
from rasa.builder.evaluator.constants import (
    DEFAULT_RESPONSE_CLASSIFICATION_EVALUATION_TEXT_OUTPUT_FILENAME,
    RESPONSE_CLASSIFICATION_EVALUATION_RESULTS_DIR,
    RESPONSE_CLASSIFICATION_EVALUATION_YAML_OUTPUT_FILENAME,
)
from rasa.builder.evaluator.copilot_executor import (
    CopilotRunResult,
    run_copilot_with_response_handler,
)
from rasa.builder.evaluator.dataset.models import DatasetEntry
from rasa.builder.evaluator.response_classification.constants import (
    EXPERIMENT_DESCRIPTION,
    EXPERIMENT_NAME,
    MACRO_F1_DESCRIPTION,
    MACRO_F1_METRIC,
    MACRO_PRECISION_DESCRIPTION,
    MACRO_PRECISION_METRIC,
    MACRO_RECALL_DESCRIPTION,
    MACRO_RECALL_METRIC,
    MICRO_F1_DESCRIPTION,
    MICRO_F1_METRIC,
    MICRO_PRECISION_DESCRIPTION,
    MICRO_PRECISION_METRIC,
    MICRO_RECALL_DESCRIPTION,
    MICRO_RECALL_METRIC,
    PER_CLASS_F1_DESCRIPTION,
    PER_CLASS_F1_METRIC_TEMPLATE,
    PER_CLASS_PRECISION_DESCRIPTION,
    PER_CLASS_PRECISION_METRIC_TEMPLATE,
    PER_CLASS_RECALL_DESCRIPTION,
    PER_CLASS_RECALL_METRIC_TEMPLATE,
    PER_CLASS_SUPPORT_DESCRIPTION,
    PER_CLASS_SUPPORT_METRIC_TEMPLATE,
    SKIP_COUNT_DESCRIPTION,
    SKIP_COUNT_METRIC,
    WEIGHTED_F1_DESCRIPTION,
    WEIGHTED_F1_METRIC,
    WEIGHTED_PRECISION_DESCRIPTION,
    WEIGHTED_PRECISION_METRIC,
    WEIGHTED_RECALL_DESCRIPTION,
    WEIGHTED_RECALL_METRIC,
)
from rasa.builder.evaluator.response_classification.evaluator import (
    ResponseClassificationEvaluator,
)
from rasa.builder.evaluator.response_classification.models import (
    ClassificationResult,
    MetricsSummary,
)

structlogger = structlog.get_logger()


class ResponseClassificationLangfuseRunner:
    """Main class for running Langfuse evaluations on the classification evaluator."""

    def __init__(self, dataset_name: str, output_dir: Optional[str] = None):
        self._langfuse = langfuse.get_client()
        self._dataset = self._retrieve_dataset(dataset_name)
        self._output_dir = (
            Path(output_dir)
            if output_dir
            else RESPONSE_CLASSIFICATION_EVALUATION_RESULTS_DIR
        )

    def _retrieve_dataset(self, dataset_name: str) -> DatasetClient:
        """Get the dataset."""
        try:
            return self._langfuse.get_dataset(dataset_name)
        except Exception as e:
            structlogger.error(
                "langfuse_runner.init.dataset_not_found",
                event_info=f"Failed to get dataset '{dataset_name}'",
                dataset_name=dataset_name,
                error=str(e),
            )
            raise

    def run_experiment(self) -> ExperimentResult:
        """Run the experiment."""
        result = self._dataset.run_experiment(
            name=EXPERIMENT_NAME,
            description=EXPERIMENT_DESCRIPTION,
            task=self._run_copilot_task,
            run_evaluators=[self._run_classification_metrics_evaluator],
        )
        self._report_run_results_to_txt_file(result)
        self._langfuse.flush()
        return result

    async def _run_copilot_task(
        self,
        *,
        item: ExperimentItem,
        **kwargs: Dict[str, Any],
    ) -> Optional[CopilotRunResult]:
        """Copilot task function that processes each dataset item.

        Follows the languse.experiment.TaskFunction protocol. The function mimics the
        functionality of the `/copilot` endpoint.

        Args:
            item: The dataset item to process.
            kwargs: Additional keyword arguments.

        Returns:
            A tuple containing the complete response and the generation context.
        """
        # Try to create the copilot context used for generating the response from the
        # dataset item, if the context cannot be created, skip the evaluation by
        # returning None.
        try:
            dataset_entry = DatasetEntry.from_raw_data(
                id=item.id,  # type: ignore[union-attr]
                input_data=item.input,  # type: ignore[union-attr]
                expected_output_data=item.expected_output,  # type: ignore[union-attr]
                metadata_data=item.metadata,  # type: ignore[union-attr]
            )
            context = dataset_entry.to_copilot_context()
        except Exception as e:
            structlogger.error(
                "langfuse_runner._task_function_run_copilot.context_creation_failed",
                event_info=(
                    f"Failed to create CopilotContext from dataset item with id: "
                    f"{item.id}. The Copilot cannot be run without a valid "  # type: ignore[union-attr]
                    f"CopilotContext. Skipping evaluation."
                ),
                item_id=item.id,  # type: ignore[union-attr]
                item_input=item.input,  # type: ignore[union-attr]
                item_expected_output=item.expected_output,  # type: ignore[union-attr]
                error=str(e),
            )
            return None

        # Run the evalution. If the task fails, skip the evaluation by returning None.
        try:
            return await run_copilot_with_response_handler(context)
        except Exception as e:
            structlogger.error(
                "langfuse_runner._task_function_run_copilot.copilot_run_failed",
                event_info=(
                    f"Failed to run the copilot with response handler for dataset item "
                    f"with id: {item.id}. Skipping evaluation."  # type: ignore[union-attr]
                ),
                item_id=item.id,  # type: ignore[union-attr]
                item_input=item.input,  # type: ignore[union-attr]
                item_expected_output=item.expected_output,  # type: ignore[union-attr]
                error=str(e),
            )
            return None

    def _run_classification_metrics_evaluator(
        self, *, item_results: List[ExperimentItemResult], **kwargs: Dict[str, Any]
    ) -> List[Evaluation]:
        """Main evaluator function that calculates classification metrics.

        This function follows the languse.experiment.RunEvaluatorFunction protocol.
        It will be called after the langfuse.experiment.TaskFunction has been called
        for each item in the dataset.

        Args:
            item_results: The item results to evaluate.
            kwargs: Additional keyword arguments.

        Returns:
            A list of Langfuse Evaluation objects.
        """
        # Create a list of ClassificationResult objects from item_results
        classification_results, skip_count = (
            self._create_classification_results_from_dataset_items(item_results)
        )

        # Log a warning if any items were skipped due to invalid data, as this will
        # affect the overall metrics.
        if skip_count > 0:
            structlogger.warning(
                "langfuse_runner._run_classification_metrics_evaluator.skipped_items",
                event_info=(
                    f"Skipped {skip_count} items due to invalid data. This will affect "
                    f"the overall metrics."
                ),
                skipped_count=skip_count,
                total_items=len(item_results),
            )

        # Run the response classification evaluator on the classification results and
        # get the metrics summary
        evaluator = ResponseClassificationEvaluator()  # type: ignore[no-untyped-call]
        metrics_summary = evaluator.evaluate(classification_results)

        # Record the metrics in Langfuse.
        evaluations = self._create_langfuse_evaluation_objects(
            metrics_summary, skip_count
        )
        self._report_yaml_structured_results(evaluations)
        return evaluations

    def _create_classification_results_from_dataset_items(
        self, item_results: List[ExperimentItemResult]
    ) -> Tuple[List[ClassificationResult], int]:
        """Create a list of ClassificationResult objects from item results.

        Args:
            item_results: The item results to create ClassificationResult objects from.

        Returns:
            A tuple containing the list of ClassificationResult objects and the number
            of items that were skipped due to missing predicted or expected categories.
        """
        classification_results: List[ClassificationResult] = []
        skip_count = 0

        # Try to create a ClassificationResult from the each item result. if either
        # predicted or expected category is missing, skip the item.
        for item_result in item_results:
            # If the output is None, the task function resulted in an error, skip the
            # item.
            if (
                item_result.output is None
                or not isinstance(item_result.output, CopilotRunResult)
                or item_result.output.response_category is None
                or item_result.item.expected_output is None  # type: ignore[union-attr]
                or not isinstance(item_result.item.expected_output, dict)  # type: ignore[union-attr]
                or item_result.item.expected_output.get("response_category") is None  # type: ignore[union-attr]
            ):
                structlogger.error(
                    "langfuse_runner._create_classification_results_from_dataset_items"
                    ".invalid_item_result",
                    event_info=(
                        f"Cannot create a ClassificationResult from item result with "
                        f"id: {item_result.item.id}. This item will not be used for "  # type: ignore[union-attr]
                        f"evaluation."
                    ),
                    item_id=item_result.item.id,  # type: ignore[union-attr]
                    item_output=item_result.output,
                    item_expected_output=item_result.item.expected_output,  # type: ignore[union-attr]
                )
                skip_count += 1
                continue

            predicted_category = item_result.output.response_category.value
            expected_category = item_result.item.expected_output["response_category"]  # type: ignore[union-attr]
            classification_result = ClassificationResult(
                prediction=ResponseCategory(predicted_category),
                expected=ResponseCategory(expected_category),
            )
            classification_results.append(classification_result)

        return classification_results, skip_count

    def _create_langfuse_evaluation_objects(
        self, metrics_summary: MetricsSummary, skip_count: int
    ) -> List[Evaluation]:
        """Create Langfuse Evaluation objects from metrics summary."""
        evaluations: List[Evaluation] = []

        # Overall metrics
        evaluations.extend(
            [
                Evaluation(
                    name=MICRO_PRECISION_METRIC,
                    value=metrics_summary.overall.micro_precision,
                    comment=MICRO_PRECISION_DESCRIPTION.format(
                        value=metrics_summary.overall.micro_precision
                    ),
                ),
                Evaluation(
                    name=MACRO_PRECISION_METRIC,
                    value=metrics_summary.overall.macro_precision,
                    comment=MACRO_PRECISION_DESCRIPTION.format(
                        value=metrics_summary.overall.macro_precision
                    ),
                ),
                Evaluation(
                    name=WEIGHTED_PRECISION_METRIC,
                    value=metrics_summary.overall.weighted_avg_precision,
                    comment=WEIGHTED_PRECISION_DESCRIPTION.format(
                        value=metrics_summary.overall.weighted_avg_precision
                    ),
                ),
                Evaluation(
                    name=MICRO_RECALL_METRIC,
                    value=metrics_summary.overall.micro_recall,
                    comment=MICRO_RECALL_DESCRIPTION.format(
                        value=metrics_summary.overall.micro_recall
                    ),
                ),
                Evaluation(
                    name=MACRO_RECALL_METRIC,
                    value=metrics_summary.overall.macro_recall,
                    comment=MACRO_RECALL_DESCRIPTION.format(
                        value=metrics_summary.overall.macro_recall
                    ),
                ),
                Evaluation(
                    name=WEIGHTED_RECALL_METRIC,
                    value=metrics_summary.overall.weighted_avg_recall,
                    comment=WEIGHTED_RECALL_DESCRIPTION.format(
                        value=metrics_summary.overall.weighted_avg_recall
                    ),
                ),
                Evaluation(
                    name=MICRO_F1_METRIC,
                    value=metrics_summary.overall.micro_f1,
                    comment=MICRO_F1_DESCRIPTION.format(
                        value=metrics_summary.overall.micro_f1
                    ),
                ),
                Evaluation(
                    name=MACRO_F1_METRIC,
                    value=metrics_summary.overall.macro_f1,
                    comment=MACRO_F1_DESCRIPTION.format(
                        value=metrics_summary.overall.macro_f1
                    ),
                ),
                Evaluation(
                    name=WEIGHTED_F1_METRIC,
                    value=metrics_summary.overall.weighted_avg_f1,
                    comment=WEIGHTED_F1_DESCRIPTION.format(
                        value=metrics_summary.overall.weighted_avg_f1
                    ),
                ),
            ]
        )

        # Per-class metrics
        for category, per_class_metrics in metrics_summary.per_class.items():
            category_name = category.value.lower()
            evaluations.extend(
                [
                    Evaluation(
                        name=PER_CLASS_PRECISION_METRIC_TEMPLATE.format(
                            category=category_name
                        ),
                        value=per_class_metrics.precision,
                        comment=PER_CLASS_PRECISION_DESCRIPTION.format(
                            category=category.value,
                            value=per_class_metrics.precision,
                        ),
                    ),
                    Evaluation(
                        name=PER_CLASS_RECALL_METRIC_TEMPLATE.format(
                            category=category_name
                        ),
                        value=per_class_metrics.recall,
                        comment=PER_CLASS_RECALL_DESCRIPTION.format(
                            category=category.value,
                            value=per_class_metrics.recall,
                        ),
                    ),
                    Evaluation(
                        name=PER_CLASS_F1_METRIC_TEMPLATE.format(
                            category=category_name
                        ),
                        value=per_class_metrics.f1,
                        comment=PER_CLASS_F1_DESCRIPTION.format(
                            category=category.value,
                            value=per_class_metrics.f1,
                        ),
                    ),
                    Evaluation(
                        name=PER_CLASS_SUPPORT_METRIC_TEMPLATE.format(
                            category=category_name
                        ),
                        value=float(per_class_metrics.support),
                        comment=PER_CLASS_SUPPORT_DESCRIPTION.format(
                            category=category.value,
                            value=per_class_metrics.support,
                        ),
                    ),
                ]
            )

        # Record the number of items that were skipped due to invalid data
        evaluations.append(
            Evaluation(
                name=SKIP_COUNT_METRIC,
                value=skip_count,
                comment=SKIP_COUNT_DESCRIPTION.format(value=skip_count),
            )
        )

        return evaluations

    def _report_run_results_to_txt_file(self, result: ExperimentResult) -> None:
        result_str = result.format().replace("\\n", "\n")
        self._output_dir.mkdir(parents=True, exist_ok=True)

        # Add timestamp prefix to filename
        current_date = datetime.now().strftime("%Y%m%d_%H%M%S")
        timestamped_filename = (
            f"{current_date}_"
            f"{DEFAULT_RESPONSE_CLASSIFICATION_EVALUATION_TEXT_OUTPUT_FILENAME}"
        )
        output_path = self._output_dir / timestamped_filename

        with open(str(output_path), "w") as f:
            f.write(result_str)
        structlogger.info(
            "langfuse_runner._report_run_results.exported",
            event_info="Evaluation results exported to text file",
            text_file=output_path,
        )

    def _report_yaml_structured_results(self, evaluations: list[Evaluation]) -> None:
        """Export evaluation results to a YAML file with structured data."""
        # Ensure results directory exists
        self._output_dir.mkdir(parents=True, exist_ok=True)

        # Add timestamp prefix to filename
        current_date = datetime.now().strftime("%Y%m%d_%H%M%S")
        timestamped_filename = (
            f"{current_date}_"
            f"{RESPONSE_CLASSIFICATION_EVALUATION_YAML_OUTPUT_FILENAME}"
        )
        output_path = self._output_dir / timestamped_filename
        # Convert evaluations to structured data
        structured_data: Dict[str, Any] = {
            "experiment": {
                "name": EXPERIMENT_NAME,
                "description": EXPERIMENT_DESCRIPTION,
                "timestamp": datetime.now().isoformat(),
            },
            "metrics": [],
        }

        # Add each evaluation as a metric
        for evaluation in evaluations:
            metric_data: Dict[str, Any] = {
                "name": evaluation.name,
                "value": evaluation.value,
                "description": evaluation.comment,
            }
            structured_data["metrics"].append(metric_data)

        # Write to YAML file
        with open(str(output_path), "w") as f:
            yaml.dump(structured_data, f, default_flow_style=False, sort_keys=False)

        structlogger.info(
            "langfuse_runner._report_yaml_structured_results.exported",
            event_info="Evaluation results exported to YAML file",
            yaml_file=output_path,
            metrics_count=len(evaluations),
        )



"""Constants for the response classification evaluator."""

from typing import List, Literal

# Options for averaging methods for Response Classification Evaluator
MICRO_AVERAGING_METHOD: Literal["micro"] = "micro"
MACRO_AVERAGING_METHOD: Literal["macro"] = "macro"
WEIGHTED_AVERAGING_METHOD: Literal["weighted"] = "weighted"

AVERAGING_METHODS: List[Literal["micro", "macro", "weighted"]] = [
    MICRO_AVERAGING_METHOD,
    MACRO_AVERAGING_METHOD,
    WEIGHTED_AVERAGING_METHOD,
]

# Overall evaluation metric names
MICRO_PRECISION_METRIC = "micro_precision"
MACRO_PRECISION_METRIC = "macro_precision"
WEIGHTED_PRECISION_METRIC = "weighted_precision"

MICRO_RECALL_METRIC = "micro_recall"
MACRO_RECALL_METRIC = "macro_recall"
WEIGHTED_RECALL_METRIC = "weighted_recall"

MICRO_F1_METRIC = "micro_f1"
MACRO_F1_METRIC = "macro_f1"
WEIGHTED_F1_METRIC = "weighted_f1"

# Skip count metric name due to invalid data
SKIP_COUNT_METRIC = "skipped_items"

# Per-class evaluation metric name templates
PER_CLASS_PRECISION_METRIC_TEMPLATE = "{category}_precision"
PER_CLASS_RECALL_METRIC_TEMPLATE = "{category}_recall"
PER_CLASS_F1_METRIC_TEMPLATE = "{category}_f1"
PER_CLASS_SUPPORT_METRIC_TEMPLATE = "{category}_support"

# Description templates for evaluation metrics
MICRO_PRECISION_DESCRIPTION = "Micro Precision: {value:.3f}"
MACRO_PRECISION_DESCRIPTION = "Macro Precision: {value:.3f}"
WEIGHTED_PRECISION_DESCRIPTION = "Weighted Precision: {value:.3f}"

MICRO_RECALL_DESCRIPTION = "Micro Recall: {value:.3f}"
MACRO_RECALL_DESCRIPTION = "Macro Recall: {value:.3f}"
WEIGHTED_RECALL_DESCRIPTION = "Weighted Recall: {value:.3f}"

MICRO_F1_DESCRIPTION = "Micro F1: {value:.3f}"
MACRO_F1_DESCRIPTION = "Macro F1: {value:.3f}"
WEIGHTED_F1_DESCRIPTION = "Weighted F1: {value:.3f}"

# Skip count metric description
SKIP_COUNT_DESCRIPTION = "Skipped {value} items due to invalid data"

# Per-class description templates
PER_CLASS_PRECISION_DESCRIPTION = "[{category}] Precision: {value:.3f}"
PER_CLASS_RECALL_DESCRIPTION = "[{category}] Recall: {value:.3f}"
PER_CLASS_F1_DESCRIPTION = "[{category}] F1: {value:.3f}"
PER_CLASS_SUPPORT_DESCRIPTION = "[{category}] Support: {value}"

# Experiment configuration
EXPERIMENT_NAME = "Copilot Response Classification Evaluation"
EXPERIMENT_DESCRIPTION = (
    "Evaluating Copilot response classification performance with per-class metrics "
    "and overall averages (micro, macro, weighted). The metric that are reported are: "
    "precision, recall, F1, support."
)
